"""
PC-BASIC - codestream.py
Code stream utilities

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from functools import partial
import io

from . import error
from . import tokens as tk
from .tokens import DIGITS, HEXDIGITS, OCTDIGITS, LETTERS
from ...compat import StreamWrapper


class CodeStream(StreamWrapper):
    """Stream of various kinds of code."""

    # whitespace
    blanks = b' \t\n'
    # line end characters for ths stream type
    end_line = None

    def __init__(self, bytesbuffer):
        """Initialise the stream."""
        self._buffer = bytesbuffer
        # we need delegation rather than inheritance to allow pickling under Python 2
        StreamWrapper.__init__(self, io.BytesIO(bytesbuffer))

    def peek(self, n=1):
        """Peek next char in stream."""
        d = self.read(n)
        self.seek(-len(d), 1)
        return d

    def skip_read(self, skip_range, n=1):
        """Skip chars in skip_range, then read next."""
        while True:
            d = self.read(1)
            # skip_range must not include ''
            if d == b'' or d not in skip_range:
                return d + self.read(n-1)

    def skip_blank_read(self, n=1):
        """Skip whitespace, then read next."""
        return self.skip_read(self.blanks, n)

    def skip_blank(self, n=1):
        """Skip whitespace, then peek next."""
        d = self.skip_read(self.blanks, n)
        self.seek(-len(d), 1)
        return d

    def backskip_blank(self):
        """Skip whitespace backwards, then peek next."""
        while True:
            self.seek(-1, 1)
            d = self.peek()
            # skip_range must not include ''
            if d == b'' or d not in self.blanks:
                return d

    def read_if(self, d, in_range):
        """Read if next char is not empty and in range."""
        if d != b'' and d in in_range:
            self.read(len(d))
            return d
        return None

    def skip_blank_read_if(self, in_range, n=1):
        """Skip whitespace, then read if next char is in range."""
        return self.read_if(self.skip_blank(n=n), in_range)

    def read_to(self, findrange):
        """Read until a character from a given range is found."""
        out = b''
        while True:
            d = self.read(1)
            if d == b'':
                break
            if d in findrange:
                break
            out += d
        self.seek(-len(d), 1)
        return out

    def require_read(self, in_range, err=error.STX):
        """Skip whitespace, read and raise error if not in range."""
        d = self.read(1)
        while d and d in self.blanks:
            d = self.read(1)
        c = d + self.read(len(in_range[0])-1)
        if not c or c not in in_range:
            self.seek(-len(c), 1)
            raise error.BASICError(err)
        return c

    # read specialised items
    # these are used for both plaintext and tokenised streams:
    # tokenised streams may contain plaintext literals and names are always plaintext

    def read_name(self):
        """Read a variable name."""
        d = self.skip_blank_read()
        if not d or d not in LETTERS:
            # variable name must start with a letter
            self.seek(-len(d), 1)
            return b''
        name = b''
        while d and d in tk.NAME_CHARS:
            name += d
            d = self.read(1)
        # only the first 40 chars are relevant in GW-BASIC, rest is discarded
        name = name[:40]
        if d in tk.SIGILS:
            name += d
        else:
            self.seek(-len(d), 1)
        # names are not case sensitive
        return name.upper()

    def read_number(self):
        """Read numeric literal."""
        c = self.peek()
        if c == b'&':
            # handle hex or oct constants
            self.read(1)
            if self.peek().upper() == b'H':
                # hex literal
                return b'&H' + self._read_hex()
            else:
                # octal literal
                return b'&O' + self._read_oct()
        elif c and c in DIGITS + b'.+-':
            # decimal literal
            return self._read_dec()
        return b''

    def _read_dec(self):
        """Read decimal literal."""
        have_exp = False
        have_point = False
        word = b''
        while True:
            c = self.read(1).upper()
            if not c:
                break
            elif c == b'.' and not have_point and not have_exp:
                have_point = True
                word += c
            elif c in b'ED' and not have_exp:
                # there's a special exception for number followed by EL or EQ
                # presumably meant to protect ELSE and maybe EQV ?
                if c == b'E' and self.peek().upper() in (b'L', b'Q'):
                    self.seek(-1, 1)
                    break
                else:
                    have_exp = True
                    word += c
            elif c in b'-+' and (not word or word[-1:] in b'ED'):
                # must be first character or in exponent
                word += c
            elif c in DIGITS + self.blanks + b'\x1c\x1d\x1f':
                # '\x1c\x1d\x1f' are ASCII separators
                # - these cause string representations to evaluate to zero
                # we'll remove blanks later but need to keep it for now
                # so we can reposition the stream on removing trailing whitespace
                word += c
            elif c in b'!#' and not have_exp:
                word += c
                # must be last character
                break
            elif c == b'%':
                # swallow a %, but break parsing
                break
            else:
                self.seek(-1, 1)
                break
        # don't claim trailing whitespace
        trimword = word.rstrip(self.blanks)
        self.seek(-len(word) + len(trimword), 1)
        # remove all internal whitespace
        word = trimword.strip(self.blanks)
        return word

    def _read_hex(self):
        """Read hexadecimal literal."""
        # pass the H in &H
        self.read(1)
        word = b''
        while True:
            c = self.peek()
            # hex literals must not be interrupted by whitespace
            if c and c in HEXDIGITS:
                word += self.read(1)
            else:
                break
        return word

    def _read_oct(self):
        """Read octal literal."""
        # O is optional, could also be &777 instead of &O777
        if self.peek().upper() == b'O':
            self.read(1)
        word = b''
        while True:
            c = self.peek()
            # oct literals may be interrupted by whitespace
            if c and c in OCTDIGITS + self.blanks:
                word += self.read(1)
            else:
                break
        return word

    def read_string(self):
        """Read a string literal."""
        word = self.read(1)
        if not word or word != b'"':
            self.seek(-len(word), 1)
            return b''
        # while tokenised numbers inside a string literal will be printed as tokenised numbers,
        # they don't actually execute as such:
        # a \00 character, even if inside a tokenised number, will break a string literal
        # and make the parser expect a line number afterwards, etc. We follow this.
        word += self.read_to((b'"',) + self.end_line)
        delim = self.read(1)
        if delim == b'"':
            word += delim
        else:
            self.seek(-len(delim), 1)
        return word


class TokenisedStream(CodeStream):
    """Stream of tokenised BASIC code."""

    end_line = tk.END_LINE

    def __init__(self, addr=None):
        """Initialise tokenised stream."""
        # memory address, if any
        self._addr = addr
        CodeStream.__init__(self, b'')

    def tell_address(self):
        """Get memory address for current stream position."""
        if self._addr is not None:
            return self._addr + self.tell()
        return None

    def skip_to(self, findrange, break_on_first_char=True):
        """Skip until character is in findrange."""
        literal = False
        rem = False
        nchars = len(findrange[0])
        while True:
            c = self.read(1)
            if c == b'':
                break
            elif c == b'"':
                literal = not literal
            elif c == tk.REM:
                rem = True
            elif c == b'\0':
                literal = False
                rem = False
            if literal or rem:
                continue
            if c + self.peek(nchars-1) in findrange:
                if break_on_first_char:
                    self.seek(-1, 1)
                    break
            break_on_first_char = True
            # not elif! if not break_on_first_char, c needs to be properly processed.
            if c == b'\0':
                # offset and line number follow
                literal = False
                off = self.read(2)
                if len(off) < 2 or off == b'\0\0':
                    break
                self.read(2)
            elif c in tk.PLUS_BYTES:
                self.read(tk.PLUS_BYTES[c])

    def skip_to_read(self, findrange):
        """Skip until character is in findrange, then read."""
        self.skip_to(findrange)
        return self.read(len(findrange[0]))

    def read_keyword_token(self):
        """Read full keyword token."""
        token = self.read(1)
        if token in (b'\xff', b'\xfe', b'\xfd'):
            token += self.read(1)
        return token

    def read_number_token(self):
        """Read full token, including trailing bytes."""
        lead = self.read(1)
        if lead not in tk.NUMBER:
            self.seek(-len(lead), 1)
            return b''
        trail = self.read(tk.PLUS_BYTES.get(lead, 0))
        return lead + trail

    def require_end(self, err=error.STX):
        """Skip whitespace, peek and raise error if not at end of statement."""
        d = self.read(1)
        while d and d in self.blanks:
            d = self.read(1)
        self.seek(-len(d), 1)
        if d not in tk.END_STATEMENT:
            raise error.BASICError(err)

    def skip_to_token(self, requested_token):
        """
        Skip over bytecode until particular token on statement start;
        this excludes statements in THEN or ELSE clauses.
        """
        while True:
            separator = self.skip_to_read(tk.END_STATEMENT)
            # skip line number, if there
            if separator == b'\0':
                # break on end of stream
                trail = self.read(4)
                if len(trail) < 4 or trail[:2] == b'\0\0':
                    return None
            # get first keyword in statement
            self.skip_blank()
            token = self.read_keyword_token()
            self.seek(-len(token), 1)
            if not token:
                return None
            elif token == requested_token:
                return token

    def skip_block(self, for_char, next_char, allow_comma=False):
        """Skip over bytecode until block end token."""
        stack = 0
        while True:
            c = self.skip_to_read(tk.END_STATEMENT + (tk.THEN, tk.ELSE))
            # skip line number, if there
            if c == b'\0':
                # break on end of stream
                trail = self.read(4)
                if len(trail) < 4 or trail[:2] == b'\0\0':
                    break
            # get first keyword in statement
            d = self.skip_blank()
            if d == b'':
                break
            elif d == for_char:
                self.read(1)
                stack += 1
            elif d == next_char:
                if stack <= 0:
                    break
                else:
                    self.read(1)
                    stack -= 1
                    # NEXT I, J
                    if allow_comma:
                        while (self.skip_blank() not in tk.END_STATEMENT):
                            self.skip_to(tk.END_STATEMENT + (b',',))
                            if self.peek() == b',':
                                if stack > 0:
                                    self.read(1)
                                    stack -= 1
                                else:
                                    return



"""
PC-BASIC - extensions.py
Extension handler

(c) 2018--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import logging
from importlib import import_module

from ..compat import text_type

from .base import error
from . import values


class Extensions(object):
    """Extension handler."""

    def __init__(self, extension, values, codepage):
        """Initialise extension handler."""
        # `extension` can be an iterable of extensions/names of extensions, just one extension,
        # or a name of an extension (as bytes or str)
        try:
            # test for being iterable
            iter(extension)
        except TypeError:
            extension = [extension]
        if isinstance(extension, (bytes, text_type)):
            extension = [extension]
        self._extension = list(extension)
        self._values = values
        self._codepage = codepage
        self._ext_funcs = None

    def __getstate__(self):
        """Pickle."""
        pickle_dict = self.__dict__.copy()
        # modules can't be pickled
        pickle_dict['_ext_funcs'] = None
        pickle_dict['step'] = None
        return pickle_dict

    def __setstate__(self, pickle_dict):
        """Unpickle."""
        self.__dict__.update(pickle_dict)
        self.step = lambda: None

    def add(self, ext):
        """Add an extension."""
        self._extension.append(ext)
        # reset cache
        self._ext_funcs = None

    def _load_extensions(self):
        """Cache extension modules and objects."""
        if self._ext_funcs is not None:
            return
        if not self._extension:
            raise error.BASICError(error.STX)
        ext_objs = []
        for ext in self._extension:
            try:
                if isinstance(ext, (bytes, text_type)):
                    ext_objs.append(import_module(ext))
                else:
                    ext_objs.append(ext)
            except Exception as e:
                logging.error(u'Could not load extension module `%s`: %s', ext, repr(e))
                raise error.BASICError(error.INTERNAL_ERROR)
        self._ext_funcs = {
            n.upper().encode('ascii', 'ignore'): getattr(ext_obj, n)
            for ext_obj in ext_objs
            for n in dir(ext_obj) if not n.startswith('_')
        }

    def call_as_statement(self, args):
        """Extension statement: call a python function as a statement."""
        self._load_extensions()
        func_name = next(args)
        func_args = list(arg.to_value() for arg in args if arg is not None)
        try:
            result = self._ext_funcs[func_name](*func_args)
        except (error.Exit, error.Reset):
            raise
        except Exception as e:
            logging.error(u'Could not call extension function `%s%s`: %s', func_name, tuple(func_args), repr(e))
            raise error.BASICError(error.INTERNAL_ERROR)
        return result

    def call_as_function(self, args):
        """Extension function: call a python function as a function."""
        result = self.call_as_statement(args)
        if isinstance(result, text_type):
            return self._values.new_string().from_str(self._codepage.unicode_to_bytes(result))
        if isinstance(result, bytes):
            return self._values.from_value(result, values.STR)
        elif isinstance(result, bool):
            return self._values.from_bool(result)
        elif isinstance(result, int) or isinstance(result, float):
            return self._values.from_value(result, values.DBL)
        raise error.BASICError(error.TYPE_MISMATCH)



"""
PC-BASIC - values package
Types, values and conversions

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from . import numbers
from . import strings
from . import values
from . import randomiser

from .numbers import *
from .strings import *
from .values import *
from .randomiser import *



"""
PC-BASIC - compat.win32
Interface for Windows system calls

(c) 2018--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os
import sys
import ctypes
import codecs
import logging
import threading
import subprocess
import msvcrt
import struct
import tempfile
import time

from ctypes.wintypes import LPCWSTR, LPWSTR, DWORD, HINSTANCE, HANDLE, HKEY, BOOL
from ctypes import cdll, windll, POINTER, pointer, c_int, c_wchar_p, c_ulonglong, byref


# text conventions
# ctrl+Z
EOF = u'\x1A'
# CRLF end-of-line
EOL = u'\r\n'


##############################################################################
# cmd.exe conventions

# register cp65001 as an alias for utf-8
codecs.register(lambda name: codecs.lookup('utf-8') if name == 'cp65001' else None)

# get OEM codepage - the one used when cmd is launched from a non-console application
HKEY_LOCAL_MACHINE = 0x80000002
KEY_QUERY_VALUE = 0x0001

def _get_oem_encoding():
    """Get Windows OEM codepage."""
    hkey = HKEY()
    windll.advapi32.RegOpenKeyExW(
        HKEY_LOCAL_MACHINE, LPWSTR(u'SYSTEM\\CurrentControlSet\\Control\\Nls\\CodePage'),
        DWORD(0), DWORD(KEY_QUERY_VALUE), byref(hkey))
    strval = ctypes.create_unicode_buffer(255)
    # key HKLM SYSTEM\\CurrentControlSet\\Control\\Nls\\CodePage value OEMCP
    size = DWORD(0)
    windll.advapi32.RegQueryValueExW(hkey, LPWSTR(u'OEMCP'), DWORD(0), None, None, byref(size))
    windll.advapi32.RegQueryValueExW(
        hkey, LPWSTR(u'OEMCP'), DWORD(0), None, byref(strval), byref(size)
    )
    windll.advapi32.RegCloseKey(hkey)
    return 'cp' + strval.value


# if starting from a console, shell will inherit its codepage
# if starting from the gui (stdin is None), we're using OEM codepage
OEM_ENCODING = _get_oem_encoding()
SHELL_ENCODING = None
if sys.stdin:
    SHELL_ENCODING = sys.stdin.encoding
SHELL_ENCODING = SHELL_ENCODING or OEM_ENCODING
# there's also an ACP codepage - this seems to be locale.getpreferredencoding()
#_ACP_ENCODING = 'cp' + str(cdll.kernel32.GetACP())


# avoid having an empty CMD window popping up in front of ours
HIDE_WINDOW = subprocess.STARTUPINFO()
HIDE_WINDOW.dwFlags |= 1  # STARTF_USESHOWWINDOW
HIDE_WINDOW.wShowWindow = 0 # SW_HIDE


##############################################################################
# various

# Windows 10 - set to DPI aware to avoid scaling twice on HiDPI screens
# see https://bitbucket.org/pygame/pygame/issues/245/wrong-resolution-unless-you-use-ctypes

try:
    set_dpi_aware = ctypes.windll.user32.SetProcessDPIAware
except AttributeError:
    # old versions of Windows don't have this in user32.dll
    def set_dpi_aware():
        """Enable HiDPI awareness."""
        pass

##############################################################################
# file system

# free space

_GetDiskFreeSpaceExW = ctypes.windll.kernel32.GetDiskFreeSpaceExW

def get_free_bytes(path):
    """Return the number of free bytes on the drive."""
    free_bytes = c_ulonglong(0)
    _GetDiskFreeSpaceExW(c_wchar_p(path), None, None, pointer(free_bytes))
    return free_bytes.value

# short file names

_GetShortPathName = ctypes.windll.kernel32.GetShortPathNameW
_GetShortPathName.argtypes = [LPCWSTR, LPWSTR, DWORD]

def get_short_pathname(native_path):
    """Return Windows short path name or None if not available."""
    try:
        length = _GetShortPathName(native_path, LPWSTR(0), DWORD(0))
        wbuffer = ctypes.create_unicode_buffer(length)
        _GetShortPathName(native_path, wbuffer, DWORD(length))
    except Exception as e:
        # something went wrong - this should be a WindowsError which is an OSError
        # but not clear
        return None
    else:
        # can also be None in wbuffer.value if error
        return wbuffer.value

# command-line arguments

_GetCommandLineW = cdll.kernel32.GetCommandLineW
_GetCommandLineW.argtypes = []
_GetCommandLineW.restype = LPCWSTR

_CommandLineToArgvW = windll.shell32.CommandLineToArgvW
_CommandLineToArgvW.argtypes = [LPCWSTR, POINTER(c_int)]
_CommandLineToArgvW.restype = POINTER(LPWSTR)

def _get_unicode_argv():
    """Convert command-line arguments to unicode."""
    # we need to go to the Windows API as argv may not be in a full unicode encoding
    # note that this will not be necessary in Python 3 where sys.argv is unicode
    # http://code.activestate.com/recipes/572200-get-sysargv-with-unicode-characters-under-windows/
    cmd = _GetCommandLineW()
    argc = c_int(0)
    argv = _CommandLineToArgvW(cmd, byref(argc))
    argv = [argv[i] for i in range(argc.value)]
    # clip off the python interpreter call, if we use it
    # anything that didn't get included in sys.argv is not for us either
    argv = argv[-len(sys.argv):]
    return argv

argv = _get_unicode_argv()


_GetFileAttributesW = windll.kernel32.GetFileAttributesW
_GetFileAttributesW.argtypes = [LPCWSTR]
_GetFileAttributesW.restype = DWORD
_FILE_ATTRIBUTE_HIDDEN = 2

def is_hidden(path):
    """File is hidden."""
    return _GetFileAttributesW(LPCWSTR(path)) & _FILE_ATTRIBUTE_HIDDEN


##############################################################################
# printing

class SHELLEXECUTEINFO(ctypes.Structure):
    _fields_ = (
        ('cbSize', DWORD),
        ('fMask', ctypes.c_ulong),
        ('hwnd', HANDLE),
        ('lpVerb', LPCWSTR),
        ('lpFile', LPCWSTR),
        ('lpParameters', LPCWSTR),
        ('lpDirectory', LPCWSTR),
        ('nShow', ctypes.c_int),
        ('hInstApp', HINSTANCE),
        ('lpIDList', ctypes.c_void_p),
        ('lpClass', LPCWSTR),
        ('hKeyClass', HKEY),
        ('dwHotKey', DWORD),
        ('hIconOrMonitor', HANDLE),
        ('hProcess', HANDLE),
    )

SEE_MASK_NOCLOSEPROCESS = 0x00000040
SEE_MASK_NOASYNC = 0x00000100

_ShellExecuteEx = ctypes.windll.shell32.ShellExecuteExW
_ShellExecuteEx.restype = BOOL
_WaitForSingleObject = ctypes.windll.kernel32.WaitForSingleObject


def get_default_printer():
    """Get the Windows default printer name."""
    try:
        _GetDefaultPrinterW = ctypes.WinDLL('winspool.drv').GetDefaultPrinterW
        length = DWORD()
        ret = _GetDefaultPrinterW(None, ctypes.byref(length))
        name = ctypes.create_unicode_buffer(length.value)
        ret = _GetDefaultPrinterW(name, ctypes.byref(length))
        return name.value
    except EnvironmentError as e:
        logging.error('Could not get default printer: %s', e)
        return u''

PRINTER_TIMEOUT_MS=1000

def _wait_for_process(handle, filename):
    """Give printing process some time to complete."""
    try:
        _WaitForSingleObject(handle, DWORD(PRINTER_TIMEOUT_MS))
    except EnvironmentError as e:
        logging.warning('Windows error: %s', e)
    # remove temporary
    os.remove(filename)

def line_print(printbuf, printer):
    """Print the buffer to a Windows printer."""
    if not printer or printer == u'default':
        printer = get_default_printer()
    if printbuf:
        with tempfile.NamedTemporaryFile(
                suffix='.txt', prefix='pcbasic-print-', delete=False) as f:
            # write UTF-8 Byte Order mark to ensure Notepad recognises encoding
            f.write(b'\xef\xbb\xbf')
            f.write(printbuf)
        sei = SHELLEXECUTEINFO()
        sei.cbSize = ctypes.sizeof(sei)
        sei.fMask = SEE_MASK_NOCLOSEPROCESS | SEE_MASK_NOASYNC
        sei.lpVerb = u'printto'
        sei.lpFile = f.name
        sei.lpParameters = u'"%s"' % printer
        sei.hProcess = HANDLE()
        try:
            _ShellExecuteEx(ctypes.byref(sei))
        except EnvironmentError as e:
            logging.error(b'Error while printing: %s', e)
        else:
            # launch non-daemon thread to wait for handle
            # to ensure we don't lose the print if triggered on exit
            threading.Thread(target=_wait_for_process, args=(sei.hProcess, f.name)).start()



"""
PC-BASIC - synthesiser.py
Tone and noise sample generator

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from math import ceil

from ..compat import xrange


# initial condition - see dosbox source
INIT_NOISE = 0x0f35
INIT_TONE = 0x01
# white noise feedback
FEEDBACK_NOISE = 0x4400
# 'periodic' feedback mask (15-bit rotation)
FEEDBACK_PERIODIC = 0x4000
# square wave feedback mask
FEEDBACK_TONE = 0x2

# number of voices
VOICES = range(4)
# voice with noise source
NOISE_VOICE = VOICES[-1]

# bit depth
SAMPLE_BITS = 8
# sample rate
SAMPLE_RATE = 44100

# The SN76489 attenuates the volume by 2dB for each step in the volume register.
# see http://www.smspower.org/Development/SN76489
# bits -2 (i.e. max div 3) so we can sum 4 voices
_MAX_AMPLITUDE = (1 << (SAMPLE_BITS-3)) - 1
# geometric list of amplitudes for volume values
# 2 dB steps correspond to a voltage factor of 10**(-2./20.) as power ~ voltage**2
_STEP_FACTOR = 10 ** (-2./20.)
# zero volume means silent
_AMPLITUDE = [0] + [int(_MAX_AMPLITUDE * _STEP_FACTOR**_power) for _power in range(14, -1, -1)]

# resolution for averaging, should be even
_RESOLUTION = 20


class SignalSource(object):
    """Linear Feedback Shift Register to generate noise or tone."""

    def __init__(self, feedback, init):
        """Initialise the signal source."""
        self._lfsr = init
        self._feedback = feedback
        # "remaining phase"/pi, i.e. runs 0 to 1 or 0 to -1 on half wavelength
        self.phase = 0.
        self.bit = 0

    def next(self):
        """Get a sample bit."""
        bit = self._lfsr & 1
        self._lfsr >>= 1
        if bit:
            self._lfsr ^= self._feedback
        self.bit = bit
        return bit


class SoundGenerator(object):
    """Sound sample chunk generator."""

    def __init__(self, signal_source, feedback, frequency, duration, loop, volume):
        """Initialise the generator."""
        # noise generator
        self._signal_source = signal_source
        self._feedback = feedback
        # actual duration and gap length
        self._duration = duration
        self._amplitude = _AMPLITUDE[volume]
        self._frequency = frequency
        self.loop = loop
        self._count_samples = 0
        self._num_samples = int(self._duration * SAMPLE_RATE)

    def build_chunk(self, length):
        """Build a sound chunk."""
        self._signal_source.feedback = self._feedback
        if self._count_samples >= self._num_samples:
            # done already
            return None
        # don't generate too many samples
        if length + self._count_samples > self._num_samples and not self.loop:
            length = (self._num_samples - self._count_samples)
        if self._frequency == 32767:
            self._frequency = 0
        # work on last element of sound queue
        if self._frequency == 0:
            chunk = bytearray(length)
        else:
            half_wavelength = SAMPLE_RATE / (2.*self._frequency)
            # generate first half-wave so as to complete the last one played
            if self._signal_source.phase:
                first_length = int(half_wavelength * self._signal_source.phase)
                first_half_wave = bytearray([self._signal_source.bit]) * first_length * _RESOLUTION
                length -= first_length
                self._signal_source.phase = 0.
            else:
                first_half_wave = bytearray()
            num_half_waves = int(ceil(length / half_wavelength))
            # generate bits
            bits = (
                self._signal_source.next()
                for _ in range(num_half_waves)
            )
            # do sampling by averaging the signal over bins of given resolution
            # this allows to use vectors all the way
            # which is *much* faster than looping over an array
            # stretch array by half_wavelength * _RESOLUTION
            stretch = int(half_wavelength * _RESOLUTION)
            waves = bytearray().join(bytearray([_b]) * stretch for _b in bits)
            matrix = bytearray().join((first_half_wave, waves))
            # cut off on round number of resolution blocks
            use_length = len(matrix) - (len(matrix) % _RESOLUTION)
            # average over blocks
            # sums are between 0 and RESOLUTION, inclusive
            sums = (
                sum(matrix[_i:_i+_RESOLUTION])
                for _i in xrange(0, use_length, _RESOLUTION)
            )
            # scale to signed amplitude
            half_res = _RESOLUTION // 2
            averages = ((_s - half_res) * self._amplitude // _RESOLUTION for _s in sums)
            # pack signed bytes into bytearray
            chunk = bytearray(_sb if _sb >= 0 else 0xff + _sb for _sb in averages)
        if not self.loop:
            # last chunk is shorter
            if self._count_samples + len(chunk) < self._num_samples:
                self._count_samples += len(chunk)
            else:
                # append final chunk
                rest_length = self._num_samples - self._count_samples
                # keep track of remaining phase to avoid ticks
                if self._frequency:
                    self._signal_source.phase = float(len(chunk) - rest_length) / half_wavelength
                else:
                    self._signal_source.phase = 0.
                chunk = chunk[:rest_length]
                # done
                self._count_samples = self._num_samples
        # if loop, attach one chunk to loop, do not increment count
        return chunk


def get_signal_sources():
    """Return three tone voices plus a noise source."""
    return [
        SignalSource(FEEDBACK_NOISE, INIT_NOISE) if _voice == NOISE_VOICE
        else SignalSource(FEEDBACK_TONE, INIT_TONE)
        for _voice in VOICES
    ]



"""
PC-BASIC - debug.py
DEBUG statement and utilities

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import io
import sys
import logging
import struct

from .basic.base import error
from .compat import PY2, text_type
from .info import get_platform_info
from .basic import values, api


class DebugException(BaseException):
    """This exception was raised deliberately through the debug module."""
    # inherit from BaseException to circumvent extension manager catching Exception

    def __repr__(self):
        return self.__doc__

    __str__ = __repr__


class DebugSession(api.Session):
    """Debugging helper."""

    def __init__(self, *args, **kwargs):
        """Initialise debugger."""
        # register as an extension
        kwargs['extension'] = tuple(kwargs.get('extension', ())) + (self,)
        api.Session.__init__(self, *args, **kwargs)

    def start(self):
        """Start the session."""
        # initialise implementation
        if api.Session.start(self):
            # replace dummy debugging step
            self.set_hook(self._debug_step)
            self._do_trace = False
            self._watch_list = []

    def _debug_step(self, token):
        """Execute traces and watches on a program step."""
        outstr = u''
        if self._do_trace:
            linum = struct.unpack_from('<H', token, 2)
            outstr += u'[%i]' % linum
        for expr in self._watch_list:
            exprstr = self.convert(expr, to_type=text_type)
            outstr += u' %s = ' % (exprstr,)
            try:
                val = self.evaluate(expr)
                print(expr, val)
                if isinstance(val, bytes):
                    outstr += u'"%s"' % self.convert(val, to_type=text_type)
                else:
                    outstr += repr(val)
            except Exception as e:
                self._handle_exception(e)
        if outstr:
            logging.debug(outstr)

    def _handle_exception(self, e):
        """Handle exception during debugging."""
        logging.debug('%s %s', type(e), str(e))
        raise

    ###########################################################################
    # debugging commands

    def dir(self):
        """Show debugging commands."""
        logging.debug('Available commands:\n' + '\n'.join(
            '    _%s: %s' % (
                n.upper(), getattr(self, n).__doc__)
                for n in dir(self)
                    if '_' not in n and callable(getattr(self, n)) and n not in dir(api.Session)
            ))

    def crash(self):
        """Simulate a crash."""
        raise DebugException()

    def python(self, cmd):
        """Execute any Python code."""
        buf = io.BytesIO() if PY2 else io.StringIO()
        save_stdout = sys.stdout
        sys.stdout = buf
        try:
            exec(cmd)
        except Exception as e:
            self._handle_exception(e)
        sys.stdout = save_stdout
        logging.debug(buf.getvalue()[:-1]) # exclude \n

    def restart(self):
        """Ctrl+Alt+Delete."""
        raise error.Reset()

    def exit(self):
        """Quit the session."""
        raise error.Exit()

    def logprint(self, *args):
        """Write arguments to log."""
        logging.debug(self.convert(b' '.join(bytes(arg) for arg in args), to_type=text_type))

    def logwrite(self, *args):
        """Write arguments to log."""
        logging.debug(' '.join(repr(arg) for arg in args))

    def trace(self, on=True):
        """Switch line number tracing on or off."""
        self._do_trace = on

    def watch(self, expr):
        """Add an expression to the watch list."""
        self._watch_list.append(expr)

    def showvariables(self):
        """Dump all variables to the log."""
        repr_vars = '\n'.join((
            '==== Scalars ='.ljust(100, '='),
            self.info.repr_scalars(),
            '==== Arrays ='.ljust(100, '='),
            self.info.repr_arrays(),
            '==== Strings ='.ljust(100, '='),
            self.info.repr_strings(),
        ))
        for s in repr_vars.split('\n'):
            logging.debug(s)

    def showscreen(self):
        """Copy the screen buffer to the log."""
        for s in self.info.repr_text_screen().split('\n'):
            logging.debug(s)

    def showprogram(self):
        """Write a marked-up hex dump of the program to the log."""
        for s in self.info.repr_program().split('\n'):
            logging.debug(s)

    def showplatform(self):
        """Show platform info."""
        logging.debug(get_platform_info())



"""
PC-BASIC - signals.py
Signals for communication between interpreter and interface

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""


class Event(object):
    """Signal object for input, video or audio queue."""

    def __init__(self, event_type, params=()):
        """Create signal."""
        self.event_type = event_type
        self.params = params

    def __repr__(self):
        """Represent signal as string."""
        return '<Event %s: %r>' % (self.event_type, self.params)


# general signals

QUIT = 'quit'


# audio queue signals

# play tone
AUDIO_TONE = 'tone'
# play noise
AUDIO_NOISE = 'noise'
# stop sound
AUDIO_STOP = 'hush'
# keep sound engine alive even if quiet
AUDIO_PERSIST = 'persist'


# video queue signals

# change video mode
VIDEO_SET_MODE = 'set_mode'
# show/hide cursor
VIDEO_SHOW_CURSOR = 'show_cursor'
# set cursor shape
VIDEO_SET_CURSOR_SHAPE = 'set_cursor_shape'
# move cursor
VIDEO_MOVE_CURSOR = 'move_cursor'
# set border attribute
VIDEO_SET_BORDER_ATTR = 'set_border_attr'
# clear rows
VIDEO_CLEAR_ROWS = 'clear_rows'
# scroll
VIDEO_SCROLL = 'scroll'
# set palette
VIDEO_SET_PALETTE = 'set_palette'
# update screen section
VIDEO_UPDATE = 'update'
# set caption message
VIDEO_SET_CAPTION = 'set_caption'
# clipboard copy reply
VIDEO_SET_CLIPBOARD_TEXT = 'set_clipboard_text'


# input queue signals

# redirect or stdio input
STREAM_CHAR = 'stream_char'
# redirect or stdio closed
STREAM_CLOSED = 'stream_closed'
# keyboard events
KEYB_DOWN = 'key_down'
KEYB_UP = 'key_up'
# light pen events
PEN_DOWN = 'pen_down'
PEN_UP = 'pen_up'
PEN_MOVED = 'pen_moved'
# joystick events
STICK_DOWN = 'stick_down'
STICK_UP = 'stick_up'
STICK_MOVED = 'stick_moved'
# clipboard events
CLIP_COPY = 'clip_copy'
CLIP_PASTE = 'clip_paste'



"""
PC-BASIC - formatter.py
Formatted output handling

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from ..base import codestream
from ..base import error
from ..base import tokens as tk
from .. import values


class Formatter(object):
    """Output string formatter."""

    def __init__(self, output, console=None):
        """Initialise."""
        self._console = console
        self._output = output

    def format(self, args):
        """PRINT: Write expressions to console or file."""
        newline = True
        for sep, value in args:
            if sep == tk.USING:
                newline = self._print_using(args)
                break
            elif sep == b',':
                self._print_comma()
            elif sep == b';':
                pass
            elif sep == tk.SPC:
                self._print_spc(values.to_int(value, unsigned=True))
            elif sep == tk.TAB:
                self._print_tab(values.to_int(value, unsigned=True))
            else:
                self._print_value(next(args))
            newline = sep not in (tk.TAB, tk.SPC, b',', b';')
        if newline:
            if self._console and self._console.overflow:
                self._output.write_line()
            self._output.write_line()

    def _print_value(self, expr):
        """Print a value."""
        # numbers always followed by a space
        if isinstance(expr, values.Number):
            word = values.to_repr(expr, leading_space=True, type_sign=False) + b' '
        else:
            word = expr.to_str()
        # output file (devices) takes care of width management;
        # we must send a whole string at a time for this to be correct.
        self._output.write(word)

    def _print_comma(self):
        """Skip to next output zone."""
        number_zones = max(1, int(self._output.width // 14))
        next_zone = int((self._output.col-1) // 14) + 1
        if next_zone >= number_zones and self._output.width >= 14 and self._output.width != 255:
            self._output.write_line()
        else:
            self._output.write(b' ' * (1 + 14*next_zone-self._output.col), can_break=False)

    def _print_spc(self, num):
        """Print SPC separator."""
        numspaces = max(0, num) % self._output.width
        self._output.write(b' ' * numspaces, can_break=False)

    def _print_tab(self, num):
        """Print TAB separator."""
        pos = max(0, num - 1) % self._output.width + 1
        if pos < self._output.col:
            self._output.write_line()
            self._output.write(b' ' * (pos-1))
        else:
            self._output.write(b' ' * (pos-self._output.col), can_break=False)

    def _print_using(self, args):
        """PRINT USING clause: Write expressions to console or file using a formatting string."""
        format_expr = values.next_string(args)
        if format_expr == b'':
            raise error.BASICError(error.IFC)
        fors = codestream.CodeStream(format_expr)
        newline, format_chars = True, False
        start_cycle = True
        initial_literal = b''
        try:
            while True:
                c = fors.peek()
                if c == b'':
                    if not format_chars:
                        # avoid infinite loop
                        break
                    # loop the format string if more variables to come
                    start_cycle = True
                    initial_literal = b''
                    fors.seek(0)
                elif c == b'_':
                    # escape char; write next char in fors or _ if this is the last char
                    if start_cycle:
                        initial_literal += fors.read(2)[-1:]
                    else:
                        self._output.write(fors.read(2)[-1:])
                else:
                    try:
                        format_field = StringField(fors)
                    except ValueError:
                        try:
                            format_field = NumberField(fors)
                        except ValueError:
                            if start_cycle:
                                initial_literal += fors.read(1)
                            else:
                                self._output.write(fors.read(1))
                            continue
                    format_chars = True
                    value = next(args)
                    if value is None:
                        newline = False
                        break
                    if start_cycle:
                        self._output.write(initial_literal)
                        start_cycle = False
                    self._output.write(format_field.format(value))
            # consume any remaining arguments / finish parser
            list(args)
        except StopIteration:
            pass
        if not format_chars:
            self._output.write(initial_literal)
            # there were no format chars in the string, illegal fn call
            raise error.BASICError(error.IFC)
        return newline


##############################################################################
# formatting functions and format string parsers

class StringField(object):
    """String Formatter for PRINT USING."""

    def __init__(self, fors):
        """Get consecutive string-related formatting tokens."""
        word = b''
        c = fors.peek()
        if c in (b'!', b'&'):
            word += fors.read(1)
        elif c == b'\\':
            word += fors.read(1)
            # count the width of the \ \ token;
            # only spaces allowed and closing \ is necessary
            while True:
                c = fors.read(1)
                word += c
                if c == b'\\':
                    break
                elif c != b' ': # can be empty as well
                    fors.seek(-len(word), 1)
                    raise ValueError()
        if not word:
            raise ValueError()
        self._string_field = word

    def format(self, value):
        """Format a string."""
        s = values.pass_string(value)
        if self._string_field == b'&':
            s = s.to_str()
        else:
            s = s.to_str().ljust(len(self._string_field))[:len(self._string_field)]
        return s

class NumberField(object):
    """Number formatter for PRINT USING."""

    def __init__(self, fors):
        """Get consecutive number-related formatting tokens."""
        word, digits_before, decimals = b'', 0, 0
        # + comes first
        leading_plus = (fors.peek() == b'+')
        if leading_plus:
            word += fors.read(1)
        # $ and * combinations
        c = fors.peek()
        if c in (b'$', b'*'):
            word += fors.read(2)
            if word[-1:] != c:
                fors.seek(-len(word), 1)
                raise ValueError()
            if c == b'*':
                digits_before += 2
                if fors.peek() == b'$':
                    word += fors.read(1)
            else:
                digits_before += 1
        # number field
        c = fors.peek()
        dot = (c == b'.')
        comma = False
        if dot:
            word += fors.read(1)
        if c in (b'.', b'#'):
            while True:
                c = fors.peek()
                if not dot and c == b'.':
                    word += fors.read(1)
                    dot = True
                elif c == b'#' or (not dot and c == b','):
                    word += fors.read(1)
                    if dot:
                        decimals += 1
                    else:
                        digits_before += 1
                        if c == b',':
                            comma = True
                else:
                    break
        if digits_before + decimals == 0:
            fors.seek(-len(word), 1)
            raise ValueError()
        # post characters
        if fors.peek(4) == b'^^^^':
            word += fors.read(4)
        if not leading_plus and fors.peek() in (b'-', b'+'):
            word += fors.read(1)
        self._tokens, self._digits_before = word, digits_before
        self._decimals, self._comma = decimals, comma

    def format(self, value):
        """Format a number to a format string."""
        value = values.pass_number(value)
        tokens = self._tokens
        digits_before = self._digits_before
        decimals = self._decimals
        comma = self._comma
        # promote ints to single
        value = value.to_float()
        # illegal function call if too many digits
        if digits_before + decimals > 24:
            raise error.BASICError(error.IFC)
        # dollar sign, decimal point
        has_dollar, force_dot = b'$' in tokens, b'.' in tokens
        # leading sign, if any
        valstr, post_sign = b'', b''
        neg = value.is_negative()
        if tokens[:1] == b'+':
            valstr += b'-' if neg else b'+'
        elif tokens[-1:] == b'+':
            post_sign = b'-' if neg else b'+'
        elif tokens[-1:] == b'-':
            post_sign = b'-' if neg else b' '
        else:
            valstr += b'-' if neg else b''
            # reserve space for sign in scientific notation by taking away a digit position
            if not has_dollar:
                digits_before -= 1
                if digits_before < 0:
                    digits_before = 0
        # take absolute value
        # NOTE: this could overflow for Integer -32768
        # but we convert to Float before calling format_number
        value = value.clone().iabs()
        # currency sign, if any
        valstr += b'$' if has_dollar else b''
        # format to string
        if b'^' in tokens:
            # don't group thousands in scientific notation
            valstr += value.to_str_scientific(digits_before, decimals, force_dot)
        else:
            valstr += value.to_str_fixed(decimals, force_dot, comma)
        # trailing signs, if any
        valstr += post_sign
        # add leading zero before radix if there's space
        if len(valstr) < len(tokens):
            if valstr.startswith(b'.'):
                valstr = b'0' + valstr
            elif valstr.startswith(b'+.'):
                valstr = b'+0' + valstr[1:]
            elif valstr.startswith(b'-.'):
                valstr = b'-0' + valstr[1:]
        if len(valstr) > len(tokens):
            # number does not fit in field
            valstr = b'%' + valstr
        else:
            # filler
            valstr = valstr.rjust(len(tokens), b'*' if b'*' in tokens else b' ')
        return valstr



"""
PC-BASIC - interface package
Video, input and audio handlers

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from .base import InitFailed, video_plugins, audio_plugins
from .interface import Interface

# video plugins
from .video import VideoPlugin
from .video_ansi import VideoANSI
from .video_cli import VideoCLI
from .video_curses import VideoCurses
from .video_pygame import VideoPygame
from .video_sdl2 import VideoSDL2

# audio plugins
from .audio import AudioPlugin
from .audio_beep import AudioBeep
from .audio_pygame import AudioPygame
from .audio_sdl2 import AudioSDL2
from .audio_portaudio import AudioPortAudio



"""
PC-BASIC - scalars.py
Scalar variable management

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import struct

from ...compat import iteritems, iterkeys

from ..base import error
from .. import values


class Scalars(object):
    """Scalar variables."""

    def __init__(self, memory, values):
        """Initialise scalars."""
        self._memory = memory
        self._values = values
        self.clear()

    def __contains__(self, varname):
        """Check if a scalar has been defined."""
        return varname in self._vars

    def __iter__(self):
        """Return an iterable over all scalar names."""
        return iterkeys(self._vars)

    def __repr__(self):
        """Debugging representation of variable dictionary."""
        return '\n'.join(
            '%s: %s' % (n.decode('ascii'), self._values.from_bytes(v))
            for n, v in iteritems(self._vars)
        )

    def clear(self):
        """Clear scalar variables."""
        self._vars = {}
        self._var_memory = {}
        self.current = 0

    @staticmethod
    def _record_size(name):
        """Calculate size of scalar record in bytes."""
        # first two bytes: chars of name or 0 if name is one byte long
        return max(3, len(name)) + 1

    @staticmethod
    def _buffer_size(name):
        """Calculate size of scalar buffer in bytes."""
        return values.size_bytes(name)

    @staticmethod
    def memory_size(name):
        """Calculate size of scalar record and buffer in bytes."""
        return Scalars._record_size(name) + Scalars._buffer_size(name)

    def set(self, name, value=None):
        """Assign a value to a variable."""
        if isinstance(value, values.String):
            self._memory.strings.fix_temporaries()
        type_char = name[-1:]
        if value is not None:
            value = values.to_type(type_char, value)
        # update memory model
        # check if garbage needs collecting before allocating memory
        if name not in self._var_memory:
            # don't add string length, string already stored
            size = self.memory_size(name)
            self._memory.check_free(size, error.OUT_OF_MEMORY)
            # first two bytes: chars of name or 0 if name is one byte long
            name_ptr = self._memory.var_current()
            # byte_size first_letter second_letter_or_nul remaining_length_or_nul
            var_ptr = name_ptr + self._record_size(name)
            self.current += size
            self._var_memory[name] = (name_ptr, var_ptr)
        # don't change the value if just checking allocation
        if value is None:
            if name in self._vars:
                return
            else:
                value = self._values.new(type_char)
        # copy buffers
        try:
            # in-place copy is crucial for FOR
            self._vars[name][:] = value.to_bytes()[:]
        except KeyError:
            # copy into new buffer if not existing
            self._vars[name] = value.to_bytes()[:]

    def get(self, name):
        """Retrieve the value of a scalar variable."""
        try:
            # we can't copy as we may end up with stale string pointers
            return self._values.create(self._vars[name])
        except KeyError:
            return self._values.new(name[-1:])

    def view(self, name):
        """Retrieve a view of an existing scalar variable."""
        return self._values.create(self._vars[name])

    def view_buffer(self, name):
        """Retrieve a view of an existing scalar variable's buffer."""
        return memoryview(self._vars[name])

    def varptr(self, name):
        """Retrieve the address of a scalar variable."""
        _, var_ptr = self._var_memory[name]
        return var_ptr

    def dereference(self, address):
        """Get a value for a scalar given its pointer address."""
        for name, data in iteritems(self._var_memory):
            if data[1] == address:
                return self.get(name)
        return None

    def get_memory(self, address):
        """Retrieve data from data memory: variable space """
        name_addr = -1
        var_addr = -1
        the_var = None
        for name in self._var_memory:
            name_try, var_try = self._var_memory[name]
            if name_try <= address and name_try > name_addr:
                name_addr, var_addr = name_try, var_try
                the_var = name
        if the_var is None: # pragma: no cover
            return -1
        if address >= var_addr:
            offset = address - var_addr
            if offset >= values.size_bytes(the_var): # pragma: no cover
                return -1
            var_rep = self._vars[the_var]
            return var_rep[offset]
        else:
            offset = address - name_addr
            return get_name_in_memory(the_var, offset)

    def get_strings(self):
        """Return a list of views of string scalars."""
        return [
            memoryview(value) for name, value in iteritems(self._vars) if name[-1:] == values.STR
        ]


###############################################################################
# variable memory

def get_name_in_memory(name, offset):
    """Memory representation of variable name."""
    # 00 type size in bytes
    # 01 1st char of name
    # 02 2nd char of name or 00
    # 03 length of name minus 3, 00 if less than 3 chars
    # 04-- remaining chars of name, excluding sigil, shifted so that A is encoded as &hC1
    #
    normname = bytearray(name.upper())[:-1]
    if offset == 0:
        return values.size_bytes(name)
    elif offset == 1:
        return normname[0]
    elif offset == 2:
        if len(name) > 2:
            return normname[1]
        else:
            return 0
    elif offset == 3:
        if len(name) > 3:
            return len(name)-3
        else:
            return 0
    elif 4 <= offset <= len(normname)+1:
        # rest of name is encoded such that c1 == 'A'
        return normname[offset-2] - ord(b'A') + 0xC1
    else: # pragma: no cover
        return -1



"""
PC-BASIC - video_ansi.py
Console text interface

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import sys

from .video import VideoPlugin
from .base import video_plugins
from . import video_cli
from ..compat import console, zip
from ..compat import iter_chunks


# CGA colours: black, cyan, magenta, white
COLOURS_4 = (0, 3, 5, 7) * 4
# Mono colours: black, white
COLOURS_2 = (0, 7) * 8


@video_plugins.register('ansi')
class VideoANSI(video_cli.VideoTextBase):
    """Text interface."""

    def __init__(self, input_queue, video_queue, caption=u'', border_width=0, **kwargs):
        """Initialise the text interface."""
        video_cli.VideoTextBase.__init__(self, input_queue, video_queue)
        # don't quit on EOF - it's counterintuitive for the text interface
        self._input_handler.quit_on_eof = False
        # window caption
        self._caption = caption
        # cursor is visible
        self._cursor_visible = True
        # cursor is block-shaped (vs line-shaped)
        self._block_cursor = False
        # current cursor position
        self._cursor_row, self._cursor_col = 1, 1
        # last used colour attributes
        self._last_attributes = None
        self._cursor_attr = None
        # text and colour buffer
        self._height, self._width = 25, 80
        self._border_y = int(round((self._height * border_width)/200.))
        self._border_x = int(round((self._width * border_width)/200.))
        self._border_attr = 0
        self.default_colours = range(16)
        self._attributes = []

    def __enter__(self):
        """Open ANSI interface."""
        video_cli.VideoTextBase.__enter__(self)
        # go into alternate screen buffer
        # stderr continues on the primary buffer
        console.start_screen()
        self.set_caption_message(u'')
        console.set_attributes(0, 0, False, False)

    def __exit__(self, type, value, traceback):
        """Close ANSI interface."""
        try:
            console.close_screen()
        finally:
            video_cli.VideoTextBase.__exit__(self, type, value, traceback)

    def _work(self):
        """Handle screen and interface events."""

    def _redraw_border(self):
        """Redraw the border."""
        if not self._border_x and not self._border_y:
            return
        self._set_attributes(0, self._border_attr, False, False)
        # draw top
        for row in range(self._border_y):
            console.move_cursor_to(row+1, 1)
            console.clear_row(self._width + 2 * self._border_x)
        # draw sides
        for row in range(self._height):
            console.move_cursor_to(row+1 + self._border_y, 1)
            console.write(u' ' * self._border_x)
            console.move_cursor_to(row+1 + self._border_y, self._width + self._border_x + 1)
            console.write(u' ' * self._border_x)
        # draw bottom
        for row in range(self._border_y):
            console.move_cursor_to(row+1 + self._border_y + self._height, 1)
            console.clear_row(self._width + 2 * self._border_x)
        console.move_cursor_to(
            self._cursor_row + self._border_y, self._cursor_col + self._border_x
        )

    def _set_default_colours(self, num_attr):
        """Set colours for default palette."""
        if num_attr == 2:
            self.default_colours = COLOURS_2
        elif num_attr == 4:
            self.default_colours = COLOURS_4
        else:
            self.default_colours = range(16)

    def _set_attributes(self, fore, back, blink, underline):
        """Set ANSI colours based on split attribute."""
        if self._last_attributes == (fore, back, blink, underline):
            return
        self._last_attributes = fore, back, blink, underline
        console.set_attributes(
            self.default_colours[fore%16], self.default_colours[back%16], blink, underline
        )

    def set_border_attr(self, attr):
        """Change border attribute."""
        if attr != self._border_attr:
            self._border_attr = attr
            self._redraw_border()

    def set_palette(self, attributes, dummy_pack_pixels):
        """Set the colour palette."""
        self._set_default_colours(len(attributes))
        rgb_table = [_fore for _fore, _, _, _ in attributes[:16]]
        if len(attributes) > 16:
            # *assume* the first 16 attributes are foreground-on-black
            # this is the usual textmode byte attribute arrangement
            fore = list(range(16)) * 16
            back = tuple(_b for _b in range(8) for _ in range(16)) * 2
        else:
            fore = list(range(len(attributes)))
            # assume black background
            # blink dim-to-bright etc won't work on terminals anyway
            back = (0,) * len(attributes)
        blink = tuple(_blink for _, _, _blink, _ in attributes)
        under = tuple(_under for _, _, _, _under in attributes)
        int_attributes = list(zip(fore, back, blink, under))
        self._attributes = int_attributes
        for index, rgb in enumerate(rgb_table):
            console.set_palette_entry(index, *rgb)

    def set_mode(self, canvas_height, canvas_width, text_height, text_width):
        """Change screen mode."""
        self._height = text_height
        self._width = text_width
        console.set_attributes(0, 0, False, False)
        console.resize(self._height + 2*self._border_y, self._width + 2*self._border_x)
        console.clear()
        self._redraw_border()
        return True

    def clear_rows(self, back_attr, start, stop):
        """Clear screen rows."""
        self._set_attributes(7, back_attr, False, False)
        for row in range(start, stop+1):
            console.move_cursor_to(row + self._border_y, 1 + self._border_x)
            console.clear_row(self._width + 2 * self._border_x)
        # draw border
        self._set_attributes(
            0, self.default_colours[self._border_attr%16], False, False
        )
        for row in range(start, stop+1):
            console.move_cursor_to(row + self._border_y, 1)
            console.write(u' ' * self._border_x)
            console.move_cursor_to(row + self._border_y, 1 + self._width + self._border_x)
            console.write(u' ' * self._border_x)
        console.move_cursor_to(
            self._cursor_row + self._border_y, self._cursor_col + self._border_x
        )

    def move_cursor(self, row, col, attr, width):
        """Move the cursor to a new position."""
        if (row, col) != (self._cursor_row, self._cursor_col):
            self._cursor_row, self._cursor_col = row, col
            console.move_cursor_to(
                self._cursor_row + self._border_y, self._cursor_col + self._border_x
            )
        # change attribute of cursor
        # cursor width is controlled by terminal
        if attr != self._cursor_attr:
            self._cursor_attr = attr
            console.set_cursor_colour(self.default_colours[attr%16])

    def show_cursor(self, cursor_on, cursor_blinks):
        """Change visibility of cursor."""
        self._cursor_visible = cursor_on
        if cursor_on:
            console.show_cursor(block=self._block_cursor)
        else:
            # force move when made visible again
            console.hide_cursor()

    def set_cursor_shape(self, from_line, to_line):
        """Set the cursor shape."""
        self._block_cursor = (to_line-from_line) >= 4
        if self._cursor_visible:
            console.show_cursor(block=self._block_cursor)

    def update(self, row, col, unicode_matrix, attr_matrix, y0, x0, sprite):
        """Put text or pixels at a given position."""
        start_col = col
        curs_row, curs_col = self._cursor_row, self._cursor_col
        for text, attrs in zip(unicode_matrix, attr_matrix):
            console.move_cursor_to(row + self._border_y, col + self._border_x)
            for unicode_list, attr in iter_chunks(text, attrs):
                fore, back, blink, underline = self._attributes[attr]
                self._set_attributes(fore, back, blink, underline)
                console.write(u''.join(unicode_list).replace(u'\0', u' '))
                col += len(unicode_list)
            row += 1
            col = start_col
        console.move_cursor_to(self._cursor_row + self._border_y, self._cursor_col + self._border_x)

    def scroll(self, direction, from_line, scroll_height, back_attr):
        """Scroll the screen between from_line and scroll_height."""
        # set the default background
        # as some (not all) consoles use the background color when inserting/deleting
        # and if they can't resize this leads to glitches outside the window
        self._set_attributes(7, 0, False, False)
        if direction == -1:
            self._scroll_up(from_line, scroll_height, back_attr)
        else:
            self._scroll_down(from_line, scroll_height, back_attr)

    def _scroll_up(self, from_line, scroll_height, back_attr):
        """Scroll the screen up between from_line and scroll_height."""
        console.scroll(from_line + self._border_y, scroll_height + self._border_y, rows=-1)
        self.clear_rows(back_attr, scroll_height, scroll_height)

    def _scroll_down(self, from_line, scroll_height, back_attr):
        """Scroll the screen down between from_line and scroll_height."""
        console.scroll(from_line + self._border_y, scroll_height + self._border_y, rows=1)
        self.clear_rows(back_attr, from_line, from_line)

    def set_caption_message(self, msg):
        """Add a message to the window caption."""
        if msg:
            console.set_caption(self._caption + u' - ' + msg)
        else:
            console.set_caption(self._caption)



"""
PC-BASIC - inputs.stick
Joystick handling

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import datetime

from ..base import error
from ..base import tokens as tk
from ..base import signals
from .. import values


class Stick(object):
    """Joystick support."""

    def __init__(self, values):
        """Initialise joysticks."""
        self._values = values
        self.is_firing = [[False, False], [False, False]]
        # axis 0--255; 128 is mid but reports 0, not 128 if no joysticks present
        self.axis = [[0, 0], [0, 0]]
        self.is_on = False
        self._was_fired = [[False, False], [False, False]]
        # timer for reading game port
        self._out_time = self._decay_timer()

    def check_input(self, signal):
        """Handle joystick-related input signals."""
        if signal.event_type == signals.STICK_DOWN:
            self._down(*signal.params)
        elif signal.event_type == signals.STICK_UP:
            self._up(*signal.params)
        elif signal.event_type == signals.STICK_MOVED:
            self._moved(*signal.params)
        else:
            return False
        return True

    def _down(self, joy, button):
        """Report a joystick button down event."""
        try:
            self._was_fired[joy][button] = True
            self.is_firing[joy][button] = True
        except IndexError:
            # ignore any joysticks/axes beyond the 2x2 supported by BASIC
            pass

    def _up(self, joy, button):
        """Report a joystick button up event."""
        try:
            self.is_firing[joy][button] = False
        except IndexError:
            # ignore any joysticks/axes beyond the 2x2 supported by BASIC
            pass

    def _moved(self, joy, axis, value):
        """Report a joystick axis move."""
        try:
            self.axis[joy][axis] = value
        except IndexError:
            # ignore any joysticks/axes beyond the 2x2 supported by BASIC
            pass

    def strig_statement_(self, args):
        """Switch joystick handling on or off."""
        on, = args
        self.is_on = (on == tk.ON)

    def stick_(self, args):
        """STICK: poll the joystick axes."""
        fn, = args
        fn = values.to_int(fn)
        error.range_check(0, 3, fn)
        joy, axis = fn // 2, fn % 2
        try:
            result = self.axis[joy][axis]
        except IndexError:
            # ignore any joysticks/axes beyond the 2x2 supported by BASIC
            result = 0
        return self._values.new_integer().from_int(result)

    def strig_(self, args):
        """STRIG: poll the joystick fire button."""
        fn, = args
        fn = values.to_int(fn)
        error.range_check(0, 7, fn)
        # [stick][button]
        # 0,1 -> [stick 0][button 0] 2,3 -> [1][0]  4,5-> [0][1]  6,7 -> [1][1]
        joy, trig = (fn//2) % 2, fn // 4
        if fn % 2 == 0:
            # has been fired
            stick_was_trig = self._was_fired[joy][trig]
            self._was_fired[joy][trig] = False
            result = -1 if stick_was_trig else 0
        else:
            # is currently firing
            result = -1 if self.is_firing[joy][trig] else 0
        return self._values.new_integer().from_int(result)

    def decay(self):
        """Return time since last game port reset."""
        return (self._decay_timer() - self._out_time) % 86400000

    def reset_decay(self):
        """Reset game port."""
        self._out_time = self._decay_timer()

    def _decay_timer(self):
        """Millisecond timer for game port decay."""
        now = datetime.datetime.now()
        return now.second*1000 + now.microsecond//1000



"""
PC-BASIC - ports.py
Serial port handling

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import logging
import sys
import os
import datetime
import io
from contextlib import contextmanager

from ...compat import iteritems
from ...compat import console, stdio

from .devicebase import safe_io

try:
    import serial
    # use the old VERSION constant as __version__ not defined in v2
    if serial.VERSION < '3':
        raise ImportError('PySerial version %s found but >= 3.0.0 required.' % serial.VERSION)
    from serial import serialutil
    logging_msg = ''
except Exception as e:
    serial = None
    logging_msg = str(e)

from ..base import error
from .. import values
from .devicebase import Device, DeviceSettings, TextFileBase, RealTimeInputMixin
from .devicebase import parse_protocol_string


###############################################################################
# COM ports

class COMDevice(Device):
    """Serial port device (COMn:)."""

    allowed_modes = b'IOAR'

    def __init__(self, arg, queues, serial_in_size):
        """Initialise COMn: device."""
        Device.__init__(self)
        # for wait()
        self._queues = queues
        self._serial_in_size = serial_in_size
        self._spec = arg
        self._serial = self._init_serial(arg)
        self.device_file = DeviceSettings()
        # only one file open at a time
        self._file = None

    def open(self, number, param, filetype, mode, access, lock, reclen, seg, offset, length, field):
        """Open a file on COMn: """
        if not self._serial:
            raise error.BASICError(error.DEVICE_UNAVAILABLE)
        # PE setting not implemented
        speed, parity, bytesize, stop, rs, cs, ds, cd, lf, _ = self._parse_params(param)
        # open the COM port
        if self._file and self._file.is_open:
            raise error.BASICError(error.FILE_ALREADY_OPEN)
        else:
            self._open_serial(rs, cs, ds, cd)
        try:
            self.set_params(speed, parity, bytesize, stop)
        except Exception:
            self.close()
            raise
        self._file = COMFile(self._serial, field, lf, self._serial_in_size, self._queues)
        # inherit width settings from device file
        # note that these seem unused for COM files
        self._file.width = self.device_file.width
        self._file.col = self.device_file.col
        return self._file

    def available(self):
        """Device is available."""
        return self._serial is not None

    def _parse_params(self, param):
        """Parse serial port connection parameters """
        max_param = 10
        param_list = param.upper().split(b',')
        if len(param_list) > max_param:
            raise error.BASICError(error.BAD_FILE_NAME)
        param_list += [b''] * (max_param-len(param_list))
        speed, parity, data, stop = param_list[:4]
        # set speed
        if speed not in (
                b'75', b'110', b'150', b'300', b'600', b'1200',
                b'1800', b'2400', b'4800', b'9600', b''
            ):
            # Bad file name
            raise error.BASICError(error.BAD_FILE_NAME)
        speed = int(speed) if speed else 300
        # set parity
        if parity not in (b'S', b'M', b'O', b'E', b'N', b''):
            raise error.BASICError(error.BAD_FILE_NAME)
        parity = parity or b'E'
        # set data bits
        if data not in (b'4', b'5', b'6', b'7', b'8', b''):
            raise error.BASICError(error.BAD_FILE_NAME)
        data = int(data) if data else 7
        bytesize = data + (parity != b'N')
        if bytesize not in range(5, 9):
            raise error.BASICError(error.BAD_FILE_NAME)
        # set stopbits
        if stop not in (b'1', b'2', b''):
            raise error.BASICError(error.BAD_FILE_NAME)
        if not stop:
            stop = 2 if (speed in (75, 110)) else 1
        else:
            stop = int(stop)
        lf, rs, cs, ds, cd, pe = False, False, None, 1000, 0, False
        for named_param in param_list[4:]:
            if not named_param:
                continue
            try:
                if named_param == b'RS':
                    # suppress request to send
                    rs = True
                elif named_param[:2] == b'CS':
                    # set CTS timeout - clear to send
                    # 0 for empty string; BAD FILE NAME if not numeric
                    cs = int(named_param[2:]) if named_param[2:] else 0
                elif named_param[:2] == b'DS':
                    # set DSR timeout - data set ready
                    ds = int(named_param[2:]) if named_param[2:] else 0
                elif named_param[:2] == b'CD':
                    # set CD timeout - carrier detect
                    cd = int(named_param[2:]) if named_param[2:] else 0
                elif named_param == b'LF':
                    # send a line feed at each return
                    lf = True
                elif named_param == b'PE':
                    # enable parity checking
                    # not implemented
                    pe = True
                else:
                    raise error.BASICError(error.BAD_FILE_NAME)
            except ValueError:
                raise error.BASICError(error.BAD_FILE_NAME)
        # CS default depends on RS
        if cs is None:
            cs = 1000 if not rs else 0
        return speed, parity, bytesize, stop, rs, cs, ds, cd, lf, pe

    def char_waiting(self):
        """Whether a char is present in buffer. For ON COM(n)."""
        if not self._serial:
            return False
        with safe_io():
            # ON COM can be set without any OPEN statement
            # so we need to ensure the serial port is opened before querying it
            if not self._serial.is_open:
                self._serial.open()
            return self._serial.in_waiting

    ##########################################################################

    def _init_serial(self, spec):
        """Initialise the serial object."""
        addr, val = parse_protocol_string(spec)
        try:
            if not addr and not val:
                pass
            elif addr == u'STDIO' or (not addr and val.upper() == u'STDIO'):
                return SerialStdIO(val.upper() == u'CRLF')
            else:
                if not serial:
                    logging.warning(
                        u'Could not attach %s to COM device. Module `serial` not available: %s',
                        spec, logging_msg
                    )
                    return None
                if addr in (u'SOCKET', u'RFC2217'):
                    # throws ValueError if too many :s, caught below
                    host, socket = val.split(u':')
                    url = u'%s://%s:%s' % (addr.lower(), host, socket)
                    stream = serial.serial_for_url(url, timeout=0, do_not_open=True)
                    # monkey-patch serial object as SocketSerial does not have this property
                    stream.out_waiting = 0
                    return stream
                elif addr == u'PORT':
                    # port can be e.g. /dev/ttyS1 on Linux or COM1 on Windows.
                    return serial.serial_for_url(val, timeout=0, do_not_open=True)
                else:
                    raise ValueError(u'Invalid protocol `%s`' % (addr,))
        except (ValueError, EnvironmentError) as e:
            logging.warning(u'Could not attach %s to COM device: %s', spec, e)
        return None

    def __getstate__(self):
        """Get pickling dict for stream."""
        # copy as we still need _serial for close()
        # which gets called after __getstate__() on shutdown
        pickle_dict = {k:v for k,v in iteritems(self.__dict__)}
        del pickle_dict['_serial']
        return pickle_dict

    def __setstate__(self, pickle_dict):
        """Initialise stream from pickling dict."""
        self.__dict__.update(pickle_dict)
        self._serial = self._init_serial(self._spec)

    def _check_open(self):
        """Open the underlying port if necessary."""
        if not self._serial.is_open:
            logging.debug('Opening serial port %s.', self._serial.port)
            self._serial.open()

    def _open_serial(self, rs=False, cs=1000, ds=1000, cd=0):
        """Open the serial connection."""
        with safe_io(error.DEVICE_TIMEOUT):
            self._check_open()
        # handshake - report as timeout if it fails
        # by default, RTS is up, DTR down
        # RTS can be suppressed, DTR only accessible through machine ports
        # https://lbpe.wikispaces.com/AccessingSerialPort
        if not rs:
            with safe_io(error.DEVICE_TIMEOUT):
                self._serial.rts = True
        now = datetime.datetime.now()
        timeout_cts = now + datetime.timedelta(microseconds=cs)
        timeout_dsr = now + datetime.timedelta(microseconds=ds)
        timeout_cd = now + datetime.timedelta(microseconds=cd)
        with safe_io(error.DEVICE_TIMEOUT):
            have_cts, have_dsr, have_cd = self._serial.cts, self._serial.dsr, self._serial.cd
        while ((now < timeout_cts and not have_cts) and
                (now < timeout_dsr and not have_dsr) and
                (now < timeout_cd and not have_cd)):
            now = datetime.datetime.now()
            with safe_io(error.DEVICE_TIMEOUT):
                have_cts = have_cts and self._serial.cts
                have_dsr = have_dsr and self._serial.dsr
                have_cts = have_cd and self._serial.cd
            # give CPU some time off
            self._queues.wait()
        # only check for status if timeouts are set > 0
        # http://www.electro-tech-online.com/threads/qbasic-serial-port-control.19286/
        # https://measurementsensors.honeywell.com/ProductDocuments/Instruments/008-0385-00.pdf
        if ((cs > 0 and not have_cts) or
                (ds > 0 and not have_dsr) or
                (cd > 0 and not have_cd)):
            raise error.BASICError(error.DEVICE_TIMEOUT)

    def set_params(self, speed, parity, bytesize, stop):
        """Set serial port connection parameters."""
        logging.debug(
            'Setting serial port %s parameters to (%d, %s, %s, %s).',
            self._serial.port, speed, parity, bytesize, stop
        )
        with safe_io(error.DEVICE_FAULT):
            self._check_open()
            self._serial.baudrate = speed
            self._serial.parity = parity.decode('ascii')
            self._serial.bytesize = bytesize
            self._serial.stopbits = stop

    def get_params(self):
        """Get serial port connection parameters."""
        with safe_io(error.DEVICE_FAULT):
            self._check_open()
            return (
                self._serial.baudrate, self._serial.parity.encode('ascii'),
                self._serial.bytesize, self._serial.stopbits
            )

    def set_pins(self, rts=None, dtr=None, brk=None):
        """Set signal pins."""
        with safe_io(error.DEVICE_FAULT):
            self._check_open()
            if rts is not None:
                self._serial.rts = rts
            if dtr is not None:
                self._serial.dtr = dtr
            if brk is not None:
                self._serial.break_condition = brk

    def get_pins(self):
        """Get signal pins."""
        with safe_io(error.DEVICE_FAULT):
            self._check_open()
            return (self._serial.cd, self._serial.ri,
                    self._serial.dsr, self._serial.cts)

    def close(self):
        """Close the serial connection."""
        if self._serial and self._serial.is_open:
            logging.debug('Closing serial port %s.', self._serial.port)
            self._serial.close()

    def io_waiting(self):
        """ Find out whether bytes are waiting for input or output. """
        # no idea what the appropriate BASIC error would be
        with safe_io(error.DEVICE_FAULT):
            self._check_open()
            # socketserial has no out_waiting, though Serial does
            return self._serial.in_waiting > 0, self._serial.out_waiting > 0


###############################################################################

class COMFile(TextFileBase, RealTimeInputMixin):
    """COMn: device - serial port."""

    def __init__(self, stream, field, linefeed, serial_in_size, queues):
        """Initialise COMn: file."""
        TextFileBase.__init__(self, stream, b'D', b'R')
        self._queues = queues
        # create a FIELD for GET and PUT. no text file operations on COMn: FIELD
        self._field = field
        self._linefeed = linefeed
        self._serial_in_size = serial_in_size
        self.is_open = True

    def close(self):
        """Close the file (but not the port)."""
        # do *not* call the parent close()
        # as this would call close() on our (unique) serial file handle
        #TextFileBase.close(self)
        self.is_open = False

    def peek(self, num):
        """Return only readahead buffer, no blocking peek."""
        return b''.join(self._readahead[:num])

    def read(self, num):
        """Read a number of characters."""
        # take at most num chars out of readahead buffer (holds just one on COM but anyway)
        s, self._readahead = self._readahead[:num], self._readahead[num:]
        while len(s) < num:
            self._queues.wait()
            with safe_io():
                # non-blocking read
                self._current, self._previous = self._fhandle.read(1), self._current
            if self._current:
                s.append(self._current)
        logging.debug('Reading from serial port %s: %r', self._fhandle.port, b''.join(s))
        return b''.join(s)

    def read_one(self):
        """Read a character, replacing CR LF with CR."""
        c = self.read(1)
        # report CRLF as CR
        # are we correct to ignore self._linefeed on input?
        if (c == b'\n' and self._previous == b'\r'):
            c = self.read(1)
        return c

    def read_line(self):
        """Blocking read line from the port (not the FIELD buffer!)."""
        out = []
        while len(out) < 255:
            c = self.read_one()
            if c == b'\r':
                break
            if c:
                out.append(c)
            c = None
        return b''.join(out), c

    def write_line(self, s=b''):
        """Write string or bytearray and newline to port."""
        self.write(s + b'\r')

    def write(self, s):
        """Write string to port."""
        if self._linefeed:
            s = s.replace(b'\r', b'\r\n')
        with safe_io():
            logging.debug('Writing to serial port %s: %r', self._fhandle.port, s)
            self._fhandle.write(s)

    def get(self, num):
        """Read num bytes - GET on COM port."""
        if not num:
            return
        # blocking read of num bytes
        s = self.read(num)
        self._field.view_buffer()[:len(s)] = s

    def put(self, num):
        """Write num bytes - PUT on COM port."""
        if not num:
            return
        self.write(bytes(self._field.view_buffer()[:num]))

    def loc(self):
        """LOC: Returns number of chars waiting to be read."""
        with safe_io():
            return self._fhandle.in_waiting

    def eof(self):
        """EOF: no chars waiting."""
        # for EOF(i)
        return self.loc() <= 0

    def lof(self):
        """Returns number of bytes free in buffer."""
        with safe_io():
            return max(0, self._serial_in_size - self._fhandle.in_waiting)


###############################################################################

class SerialStdIO(object):
    """Wrapper object to route port to stdio."""

    # dummy input pins
    cd = True
    ri = False
    dsr = True
    cts = True

    def __init__(self, crlf):
        """Initialise the stream."""
        self.is_open = False
        self._crlf = crlf
        # dummy parameters
        self.baudrate = 300
        self.parity = 'E'
        self.bytesize = 8
        self.stopbits = 2
        # dummy output pins
        self.rts = False
        self.dtr = False
        self.break_condition = False
        self.port = u'STDIO'

    def open(self):
        """Open a connection."""
        self.is_open = True

    def close(self):
        """Close the connection."""
        self.is_open = False

    def read(self, num=1):
        """Non-blocking read of up to `num` chars from stdin."""
        s = []
        # note that kbhit assumes keyboard
        # so won't work with redirects on Windows
        while console.key_pressed() and len(s) < num:
            c = stdio.stdin.buffer.read(1)
            if self._crlf and c == b'\n':
                c = b'\r'
            if c:
                s.append(c)
        return b''.join(s)

    def write(self, s):
        """Write to stdout."""
        if self._crlf:
            s = s.replace(b'\r', b'\n')
        stdio.stdout.buffer.write(s)
        stdio.stdout.buffer.flush()

    @property
    def in_waiting(self):
        """Number of characters waiting to be read."""
        # we get at most 1 char waiting this way
        return console.key_pressed()

    out_waiting = 0



"""
PC-BASIC - scancode.py
Keyboard scancodes

(c) 2014--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

# these are PC-XT keyboard scancodes without E0 extended codes
# they are labelled by their value on an IBM PC/XT US keyboard
# but they correspond to a physical key regardless of its logical value
# http://www.quadibloc.com/comp/scan.htm
#
# ROW 0
ESCAPE = 0x01
N1 = 0x02
N2 = 0x03
N3 = 0x04
N4 = 0x05
N5 = 0x06
N6 = 0x07
N7 = 0x08
N8 = 0x09
N9 = 0x0A
N0 = 0x0B
MINUS = 0x0C
EQUALS = 0x0D
BACKSPACE = 0x0E
# ROW 1
TAB = 0x0F
q = 0x10
w = 0x11
e = 0x12
r = 0x13
t = 0x14
y = 0x15
u = 0x16
i = 0x17
o = 0x18
p = 0x19
LEFTBRACKET = 0x1A
RIGHTBRACKET = 0x1B
RETURN = 0x1C
# ROW 2
CTRL = 0x1D
a = 0x1E
s = 0x1F
d = 0x20
f = 0x21
g = 0x22
h = 0x23
j = 0x24
k = 0x25
l = 0x26
SEMICOLON = 0x27
QUOTE = 0x28
BACKQUOTE = 0x29
# ROW 3
LSHIFT = 0x2A
BACKSLASH = 0x2B
z = 0x2C
x = 0x2D
c = 0x2E
v = 0x2F
b = 0x30
n = 0x31
m = 0x32
COMMA = 0x33
PERIOD = 0x34
SLASH = 0x35
RSHIFT = 0x36
PRINT = 0x37; KPTIMES = 0x37;
# ROW 4
ALT = 0x38
SPACE = 0x39
CAPSLOCK = 0x3A
# function keypad
F1 = 0x3B
F2 = 0x3C
F3 = 0x3D
F4 = 0x3E
F5 = 0x3F
F6 = 0x40
F7 = 0x41
F8 = 0x42
F9 = 0x43
F10 = 0x44
# numeric keypad
NUMLOCK = 0x45
SCROLLOCK = 0x46; BREAK = 0x46; PAUSE = 0x46;
KP7 = 0x47; HOME = 0x47
KP8 = 0x48; UP = 0x48
KP9 = 0x49; PAGEUP = 0x49
KPMINUS = 0x4A
KP4 = 0x4B; LEFT = 0x4B
KP5 = 0x4C
KP6 = 0x4D; RIGHT = 0x4D
KPPLUS = 0x4E
KP1 = 0x4F; END = 0x4F
KP2 = 0x50; DOWN = 0x50
KP3 = 0x51; PAGEDOWN = 0x51
KP0 = 0x52; INSERT = 0x52
KPPOINT = 0x53; DELETE = 0x53
# extensions (keys not on the US PC/XT keyboard)
SYSREQ = 0x54
# 56 - INT1, next to left shift, (\ on UK keyboard)
INT1 = 0x56
F11 = 0x57
F12 = 0x58
# Windows or Logo keys
LSUPER = 0x5B
RSUPER = 0x5C
MENU = 0x5D

# Japanese and Brazilian keyboards
# https://www.win.tue.nl/~aeb/linux/kbd/scancodes-8.html
# https://en.wikipedia.org/wiki/Language_input_keys
HIRAGANA_KATAKANA = 0x70
# backslash/underscore on Japanese keyboard
# /? on Brazilian keyboard
INT3 = 0x73
HENKAN = 0x79
MUHENKAN = 0x7B
ZENKAKU_HANKAKU = 0x29
# yen key on Japanese keyboard
INT4 = 0x7D
# keypad . on Brazilian keyboard
INT5 = 0x7E

# Korean keyboards
# https://www.win.tue.nl/~aeb/linux/kbd/scancodes-9.html
# Hanja key (left of Space on Korean keyboard)
HANJA = 0xF1
# Hangul/English key (right of Space on Korean keyboard)
HAN_YEONG = 0xF2


# tandy scancodes
# origin unclear, these came in with commit ce921e6 on 2014-08-18
# they may have been found experimentally on DOSBox
#F11 = 0xF9
#F12 = 0xFA

# Tandy 1000 scancodes per tandy_1k_basic_guide.pdf
# also https://manualzz.com/doc/o/144gt5/tandy-1000-ms-dos-user-manual-keyboard-ascii-and-scan-codes
# these scancodes are generally different to PC-XT from 0x29 onward
#UP = 0x29
#LEFT = 0x2B
#BACKSLASH = 0x47
#RIGHT = 0x4E
#BREAK = 0x54
#KPPLUS = 0x55; INSERT = 0x55
#KPPOINT = 0x56
#KPRETURN = 0x57 # this is not a defined constant in the XT list
#HOME = 0x58
#F11 = 0x59
#F12 = 0x5A



"""
PC-BASIC - display.textscreen
Text operations

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import logging
from contextlib import contextmanager

from ...compat import iterchar, text_type

from ..base import error
from ..base import tokens as tk
from ..base.tokens import ALPHANUMERIC
from .. import values


class ScrollArea(object):
    """Text viewport / scroll area."""

    def __init__(self, mode):
        """Initialise the scroll area."""
        self._height = mode.height
        self.unset()

    def init_mode(self, mode):
        """Initialise the scroll area for new screen mode."""
        self._height = mode.height
        if self._bottom == self._height:
            # tandy/pcjr special case: VIEW PRINT to 25 is preserved
            self.set(1, self._height)
        else:
            self.unset()

    @property
    def active(self):
        """A viewport has been set."""
        return self._active

    @property
    def bounds(self):
        """Return viewport bounds."""
        return self._top, self._bottom

    @property
    def top(self):
        """Return viewport top bound."""
        return self._top

    @property
    def bottom(self):
        """Return viewport bottom bound."""
        return self._bottom

    def set(self, start, stop):
        """Set the scroll area."""
        self._active = True
        # _top and _bottom are inclusive and count rows from 1
        self._top = start
        self._bottom = stop

    def unset(self):
        """Unset scroll area."""
        # there is only one VIEW PRINT setting across all pages.
        # scroll area normally excludes the bottom bar
        self.set(1, self._height - 1)
        self._active = False


class BottomBar(object):
    """Key guide bar at bottom line."""

    def __init__(self):
        """Initialise bottom bar."""
        # use 80 here independent of screen width
        # we store everything in a buffer and only show what fits
        self.clear()
        self.visible = False

    def clear(self):
        """Clear the contents."""
        self._contents = [(b' ', 0)] * 80

    def write(self, s, col, reverse):
        """Write chars on virtual bottom bar."""
        for i, c in enumerate(iterchar(s)):
            self._contents[col + i] = (c, reverse)

    def get_char_reverse(self, col):
        """Retrieve char and reverse attribute."""
        return self._contents[col]


class TextScreen(object):
    """Text screen."""

    def __init__(self, values, mode, cursor, capabilities):
        """Initialise text-related members."""
        self._values = values
        self._tandytext = capabilities in ('pcjr', 'tandy')
        # cursor
        self._cursor = cursor
        # current row and column
        # overflow: true if we're on 80 but should be on 81
        self.current_row, self.current_col, self.overflow = 1, 1, False
        # text viewport parameters
        self.scroll_area = ScrollArea(mode)
        # writing on bottom row is allowed
        self._bottom_row_allowed = False
        # function key macros
        self._bottom_bar = BottomBar()
        # initialised by init_mode
        self.mode = None
        self._attr = 0
        self._apagenum = 0
        self._vpagenum = 0
        self._pages = None
        self._apage = None
        self._locked = False

    def init_mode(
            self, mode, pages, attr, vpagenum, apagenum,
        ):
        """Reset the text screen for new video mode."""
        self.mode = mode
        self._attr = attr
        self._apagenum = apagenum
        self._vpagenum = vpagenum
        # character buffers
        self._pages = pages
        # pixel buffer
        self._apage = self._pages[self._apagenum]
        # redraw key line
        self.redraw_bar()
        # initialise text viewport & move cursor home
        self.scroll_area.init_mode(self.mode)
        self.set_pos(self.scroll_area.top, 1)

    def __repr__(self):
        """Return an ascii representation of the screen buffer (for debugging)."""
        return '\n'.join(repr(page) for page in self._pages)

    def set_page(self, vpagenum, apagenum):
        """Set visible and active page."""
        self._vpagenum = vpagenum
        self._apagenum = apagenum
        self._apage = self._pages[self._apagenum]
        # cursor lives on active page
        # this is technically only the case in graphics mode -
        # in DOSBox it's visible in text mode if the active page is not visible
        # but then the cursor location is static and does not equal the text insert location
        # so this seems acceptable
        self._cursor.set_active(self._vpagenum == self._apagenum)

    def set_attr(self, attr):
        """Set attribute."""
        self._attr = attr


    ###########################################################################
    # basic text buffer operations

    def write_chars(self, chars, do_scroll_down):
        """Put one character at the current position."""
        with self.collect_updates():
            for char in iterchar(chars):
                self.write_char(char, do_scroll_down)

    def write_char(self, char, do_scroll_down):
        """Put one character at the current position."""
        # see if we need to wrap and scroll down
        self._consume_overflow_before_write(do_scroll_down)
        # move cursor and see if we need to scroll up
        self._wrap_around_and_scroll_as_needed(scroll_ok=True)
        self._refresh_cursor()
        # put the character
        self._apage.put_char_attr(
            self.current_row, self.current_col, char, self._attr, adjust_end=True
        )
        # move cursor. if on col 80, only move cursor to the next row
        # when the char is printed, except if the row already wraps into the next one
        if self.current_col < self.mode.width:
            self.current_col += 1
        elif self.wraps(self.current_row):
            self.current_row += 1
            self.current_col = 1
        else:
            self.overflow = True
        # move cursor and see if we need to scroll up
        self._wrap_around_and_scroll_as_needed(scroll_ok=True)
        self._refresh_cursor()

    def _consume_overflow_before_write(self, do_scroll_down):
        """Move from overflow position to next line and set wrap flag, scroll down if needed."""
        if self.overflow:
            self.current_col += 1
            self.overflow = False
        if self.current_col > self.mode.width:
            if self.current_row < self.mode.height:
                if not self.wraps(self.current_row):
                    if do_scroll_down:
                        # scroll down (make space by shifting the next rows down)
                        if self.current_row < self.scroll_area.bottom:
                            self.scroll_down(self.current_row+1)
                    # wrap line
                    self.set_wrap(self.current_row, True)
                # move cursor and reset cursor attribute
                self.current_row, self.current_col = self.current_row + 1, 1
                self._refresh_cursor()
            else:
                self.current_col = self.mode.width

    def set_wrap(self, row, wrap):
        """Connect/disconnect rows on active page by line wrap."""
        self._apage.set_wrap(row, wrap)

    def wraps(self, row):
        """The given row is connected by line wrap."""
        return self._apage.wraps(row)

    def set_row_length(self, row, length):
        """Set logical length of row."""
        self._apage.set_row_length(row, length)

    def row_length(self, row):
        """Return logical length of row."""
        return self._apage.row_length(row)


    ###########################################################################
    # cursor position

    def up(self):
        """Move the current position 1 row up."""
        self.set_pos(self.current_row - 1, self.current_col, scroll_ok=False)

    def down(self):
        """Move the current position 1 row down."""
        self.set_pos(self.current_row + 1, self.current_col, scroll_ok=False)

    def incr_pos(self, force_halfwidth=False):
        """Increase the current position by a char width."""
        if self.overflow or force_halfwidth:
            # if we're in overflow, there's no character yet. So it's halfwidth by default.
            step = 1
        else:
            step = self._apage.get_charwidth(self.current_row, self.current_col)
            # on a trail byte: go just one to the right
            step = step or 1
        self.set_pos(self.current_row, self.current_col + step, scroll_ok=False)

    def decr_pos(self):
        """Decrease the current position by a char width."""
        # apply overflow to column number
        if self.overflow:
            self.current_col += 1
            self.overflow = False
        # check width of cell to the left
        width = self._apage.get_charwidth(self.current_row, self.current_col-1)
        # previous is trail byte: go two to the left
        # lead byte: go three to the left
        if width == 0:
            step = 2
        elif width == 2:
            step = 3
        else:
            step = 1
        self.set_pos(self.current_row, self.current_col - step, scroll_ok=False)

    def move_to_end(self):
        """Jump to end of logical line; follow wraps (END)."""
        row = self.find_end_of_line(self.current_row)
        if self.row_length(row) == self.mode.width:
            self.set_pos(row, self.row_length(row))
            self.overflow = True
        else:
            self.set_pos(row, self.row_length(row) + 1)

    def set_pos(self, to_row, to_col, scroll_ok=True):
        """Set the current position."""
        # overflow status is maintained under up or down movements
        # e.g. if we print a char while on col 80, we end up in overflow position.
        # down-error from there gets us to overflow position on the next row.
        if to_col < self.mode.width:
            self.overflow = False
        self.current_row, self.current_col = to_row, to_col
        # move cursor and reset cursor attribute
        # this may alter self.current_row, self.current_col
        self._wrap_around_and_scroll_as_needed(scroll_ok)
        self._refresh_cursor()

    def _wrap_around_and_scroll_as_needed(self, scroll_ok):
        """Check if we have crossed the screen boundaries and move as needed."""
        if self._bottom_row_allowed:
            if self.current_row == self.mode.height:
                self.current_col = min(self.mode.width, self.current_col)
                if self.current_col < 1:
                    self.current_col += 1
                return
            else:
                # if row > height, we also end up here
                # (eg if we do INPUT on the bottom row)
                # adjust viewport if necessary
                self._bottom_row_allowed = False
        # see if we need to move to the next row
        if self.current_col > self.mode.width:
            if self.current_row < self.scroll_area.bottom or scroll_ok:
                # either we don't need to scroll, or we're allowed to
                self.current_col -= self.mode.width
                self.current_row += 1
            else:
                # we can't scroll, so we just stop at the right border
                self.current_col = self.mode.width
        # see if we need to move a row up
        elif self.current_col < 1:
            if self.current_row > self.scroll_area.top:
                self.current_col += self.mode.width
                self.current_row -= 1
            else:
                self.current_col = 1
        # see if we need to scroll
        if self.current_row > self.scroll_area.bottom:
            if scroll_ok:
                self.scroll()
            self.current_row = self.scroll_area.bottom
        elif self.current_row < self.scroll_area.top:
            self.current_row = self.scroll_area.top

    @contextmanager
    def collect_updates(self):
        """Lock cursor to collect updates and submit them in one go."""
        save, self._locked = self._locked, True
        try:
            with self._apage.collect_updates():
                yield
        finally:
            self._locked = save
            self._refresh_cursor()

    def _refresh_cursor(self):
        """Move the cursor to the current position and update its attributes."""
        if self._locked:
            return
        row, col = self.current_row, self.current_col
        # in text mode, set the cursor width and attriute to that of the new location
        if self.mode.is_text_mode:
            # set halfwidth/fullwidth cursor
            width = self._apage.get_charwidth(row, col)
            # set the cursor attribute
            attr = self._apage.get_attr(row, col)
            self._cursor.move(row, col, attr, width)
        else:
            # move the cursor
            self._cursor.move(row, col)

    ###########################################################################
    # clearing the screen

    def clear_view(self):
        """Clear the scroll area."""
        with self._modify_attr_on_clear():
            self._apage.clear_rows(self.scroll_area.top, self.scroll_area.bottom, self._attr)
            self.set_pos(self.scroll_area.top, 1)

    def clear(self):
        """Clear the screen."""
        with self._modify_attr_on_clear():
            self._apage.clear_rows(1, self.mode.height, self._attr)
            self.set_pos(1, 1)

    @contextmanager
    def _modify_attr_on_clear(self):
        """On some adapters, modify character attributes when clearing the scroll area."""
        if not self._tandytext:
            # keep background, set foreground to 7
            attr_save = self._attr
            self.set_attr(attr_save & 0x70 | 0x7)
            yield
            self.set_attr(attr_save)
        else:
            yield


    ###########################################################################
    # scrolling

    def scroll(self, from_row=None):
        """Scroll the scroll region up by one row, starting at from_row."""
        if from_row is None:
            from_row = self.scroll_area.top
        self._apage.scroll_up(from_row, self.scroll_area.bottom, self._attr)
        if self.current_row > from_row:
            self.current_row -= 1
            self._refresh_cursor()


    def scroll_down(self, from_row):
        """Scroll the scroll region down by one row, starting at from_row."""
        self._apage.scroll_down(from_row, self.scroll_area.bottom, self._attr)
        if self.current_row >= from_row:
            self.current_row += 1
            self._refresh_cursor()


    ###########################################################################
    # console operations

    # logical line

    def find_start_of_line(self, srow):
        """Find the start of the logical line that includes our current position."""
        # move up as long as previous line wraps
        while srow > 1 and self._apage.wraps(srow-1):
            srow -= 1
        return srow

    def find_end_of_line(self, srow):
        """Find the end of the logical line that includes our current position."""
        # move down as long as this line wraps
        while srow <= self.mode.height and self._apage.wraps(srow):
            srow += 1
        return srow

    def find_position_in_line(self, start_row, position):
        """Set the cursor to given position in the logical line, following wraps and row length."""
        row = self.find_start_of_line(start_row)
        col = position + 1
        while (
                position >= self.row_length(row)
                and row <= self.mode.height and self._apage.wraps(row)
            ):
            col -= self.row_length(row)
            row += 1
        if col >= self.row_length(row):
            col = self.row_length(row) - 1
        return row, col

    # delete

    def delete_fullchar(self):
        """Delete the character (half/fullwidth) at the current position."""
        width = self._apage.get_charwidth(self.current_row, self.current_col)
        # on a halfwidth char, delete once; lead byte, delete twice; trail byte, do nothing
        with self.collect_updates():
            if width > 0:
                self._delete_at(self.current_row, self.current_col)
            if width == 2:
                self._delete_at(self.current_row, self.current_col)

    def _delete_at(self, row, col, remove_depleted=False):
        """Delete the halfwidth character at the given position."""
        # case 0) non-wrapping row:
        #           0a) left of or at logical end -> redraw until logical end
        #           0b) beyond logical end -> do nothing
        # case 1) full wrapping row -> redraw until physical end -> recurse for next row
        # case 2) LF row:
        #           2a) left of LF logical end ->  redraw until logical end
        #           2b) at or beyond LF logical end
        #                   -> attach next row's contents at current postion until physical end
        #                   -> if next row now empty, scroll it up & stop; otherwise recurse
        # note that the last line recurses into a multi-character delete!
        if not self.wraps(row):
            # case 0b
            if col > self.row_length(row):
                return
            # case 0a
            self._apage.delete_char_attr(row, col, self._attr)
            # if the row is depleted, drop it and scroll up from below
            if remove_depleted and self.row_length(row) == 0:
                self.scroll(row)
        elif self.row_length(row) == self.mode.width:
            # case 1
            wrap_char_attr = (
                self._apage.get_char(row+1, 1),
                self._apage.get_attr(row+1, 1)
            )
            if self.row_length(row + 1) == 0:
                wrap_char_attr = None
            self._apage.delete_char_attr(
                row, col, self._attr, wrap_char_attr
            )
            self._delete_at(row+1, 1, remove_depleted=True)
        elif col < self.row_length(row):
            # case 2a
            self._apage.delete_char_attr(row, col, self._attr)
        elif remove_depleted and col == self.row_length(row):
            # case 2b (ii) while on the first LF row deleting the last char immediately appends
            # the next row, any subsequent LF rows are only removed once they are fully empty and
            # DEL is pressed another time
            self._apage.delete_char_attr(row, col, self._attr)
        elif remove_depleted and self.row_length(row) == 0:
            # case 2b (iii) this is where the empty row mentioned at 2b (ii) gets removed
            self.scroll(row)
            return
        else:
            # case 2b (i) perform multi_character delete by looping single chars
            for newcol in range(col, self.mode.width+1):
                if self.row_length(row + 1) == 0:
                    break
                wrap_char = self._apage.get_char(row+1, 0)
                self._apage.put_char_attr(row, newcol, wrap_char, self._attr, adjust_end=True)
                self._delete_at(row+1, 1, remove_depleted=True)

    # insert

    def insert_fullchars(self, sequence):
        """Insert one or more half- or fullwidth characters and adjust cursor."""
        # insert one halfwidth character at a time at cursor location
        # to let cursor position logic deal with scrolling
        with self.collect_updates():
            for c in iterchar(sequence):
                if self._insert_at(self.current_row, self.current_col, c, self._attr):
                    # move cursor by one halfwidth character
                    # this will move to next row when necessary
                    self.incr_pos(force_halfwidth=True)

    def _insert_at(self, row, col, c, attr):
        """Insert one halfwidth character at the given position."""
        if self.row_length(row) < self.mode.width:
            # insert the new char and ignore what drops off at the end
            # this changes the attribute of everything that has been redrawn
            self._apage.insert_char_attr(row, col, c, attr)
            # the insert has now filled the row and we used to be a row ending in LF:
            # scroll and continue into the new row
            if self.wraps(row) and self.row_length(row) == self.mode.width:
                # since we used to be an LF row, wrap == True already
                # then, the newly added row should wrap - TextBuffer.scroll_down takes care of this
                self.scroll_down(row+1)
            # if we filled out the row but aren't wrapping, we scroll & wrap at the *next* insert
            return True
        else:
            # we have therow.end == width, so we're pushing the end of the row past the screen edge
            # if we're not a wrapping line, make space by scrolling and wrap into the new line
            if not self.wraps(row) and row < self.scroll_area.bottom:
                self.scroll_down(row+1)
                self.set_wrap(row, True)
            if row >= self.scroll_area.bottom:
                # once the end of the line hits the bottom, start scrolling the start of the line up
                start = self.find_start_of_line(self.current_row)
                # if we hist the top of the screen, stop inserting & drop chars
                if start <= self.scroll_area.top:
                    return False
                # scroll up
                self.scroll()
                # adjust working row number
                row -= 1
            popped_char = self._apage.insert_char_attr(row, col, c, attr)
            # insert the character in the next row
            return self._insert_at(row+1, 1, popped_char, attr)

    # line feed

    def line_feed(self):
        """Move the remainder of the line to the next row and wrap (LF)."""
        if self.current_col < self.row_length(self.current_row):
            # insert characters, preserving cursor position
            cursor = self.current_row, self.current_col
            self.insert_fullchars(b' ' * (self.mode.width-self.current_col+1))
            self.set_pos(*cursor, scroll_ok=False)
            # adjust end of line and wrapping flag - LF connects lines like word wrap
            self.set_row_length(self.current_row, self.current_col - 1)
            self.set_wrap(self.current_row, True)
            # cursor stays in place after line feed!
        else:
            # find last row in logical line
            end = self.find_end_of_line(self.current_row)
            # if the logical line hits the bottom, start scrolling up to make space...
            if end >= self.scroll_area.bottom:
                # ... until the it also hits the top; then do nothing
                start = self.find_start_of_line(self.current_row)
                if start > self.scroll_area.top:
                    self.scroll()
                else:
                    return
            # self.current_row has changed, don't use row var
            if self.current_row < self.mode.height:
                self.scroll_down(self.current_row+1)
            # ensure the current row now wraps
            self.set_wrap(self.current_row, True)
            # cursor moves to start of next line
            self.set_pos(self.current_row+1, 1)

    # console calls

    def clear_line(self, start_row, start_col, quirky_scrolling=False):
        """Clear from given position to end of logical line (CTRL+END, ESC)."""
        end_row = self.find_end_of_line(start_row)
        # clear the first row of the logical line
        self._apage.clear_row_from(start_row, start_col, self._attr)
        # input anomaly: when interacting with INPUT and ESC is pressed,
        # the first line gets cleared but not scrolled
        if quirky_scrolling and end_row > start_row:
            self._apage.clear_row_from(start_row+1, 1, self._attr)
            scroll_row = start_row + 1
        else:
            scroll_row = start_row
        # remove the additional rows in the logical line by scrolling up
        for row in range(end_row, scroll_row, -1):
            self.scroll(row)
        self.set_pos(start_row, start_col)

    def newline(self, wrap):
        """Write a newline with or without wrap."""
        self.set_wrap(self.current_row, wrap)
        self.set_pos(self.current_row + 1, 1, scroll_ok=True)

    def backspace(self, prompt_row, furthest_left):
        """Delete the char to the left (BACKSPACE)."""
        row, col = self.current_row, self.current_col
        start_row = self.find_start_of_line(row)
        # don't backspace through prompt or through start of logical line
        # on the prompt row, don't go any further back than we've been already
        if (
                ((col != furthest_left or row != prompt_row)
                and (col > 1 or row > start_row))
            ):
            self.decr_pos()
        self.delete_fullchar()

    def tab(self, overwrite):
        """Jump to next 8-position tab stop (TAB)."""
        newcol = 9 + 8 * int((self.current_col-1) // 8)
        if overwrite:
            self.set_pos(self.current_row, newcol, scroll_ok=False)
        else:
            self.insert_fullchars(b' ' * (newcol-self.current_col))

    def skip_word_right(self):
        """Skip one word to the right (CTRL+RIGHT)."""
        crow, ccol = self.current_row, self.current_col
        # find non-alphanumeric chars
        while True:
            c = self._apage.get_char(crow, ccol)
            if (c not in ALPHANUMERIC):
                break
            ccol += 1
            if ccol > self.mode.width:
                if crow >= self.scroll_area.bottom:
                    # nothing found
                    return
                crow += 1
                ccol = 1
        # find alphanumeric chars
        while True:
            c = self._apage.get_char(crow, ccol)
            if (c in ALPHANUMERIC):
                break
            ccol += 1
            if ccol > self.mode.width:
                if crow >= self.scroll_area.bottom:
                    # nothing found
                    return
                crow += 1
                ccol = 1
        self.set_pos(crow, ccol)

    def skip_word_left(self):
        """Skip one word to the left (CTRL+LEFT)."""
        crow, ccol = self.current_row, self.current_col
        # find alphanumeric chars
        while True:
            ccol -= 1
            if ccol < 1:
                if crow <= self.scroll_area.top:
                    # not found
                    return
                crow -= 1
                ccol = self.mode.width
            c = self._apage.get_char(crow, ccol)
            if (c in ALPHANUMERIC):
                break
        # find non-alphanumeric chars
        while True:
            last_row, last_col = crow, ccol
            ccol -= 1
            if ccol < 1:
                if crow <= self.scroll_area.top:
                    break
                crow -= 1
                ccol = self.mode.width
            c = self._apage.get_char(crow, ccol)
            if (c not in ALPHANUMERIC):
                break
        self.set_pos(last_row, last_col)

    ###########################################################################
    # bottom bar

    def update_bar(self, descriptions):
        """Update the key descriptions in the bottom bar."""
        self._bottom_bar.clear()
        for i, text in enumerate(descriptions):
            kcol = 1 + 8*i
            self._bottom_bar.write((b'%d' % (i+1,))[-1:], kcol, False)
            self._bottom_bar.write(text, kcol+1, True)

    def show_bar(self, on):
        """Switch bottom bar visibility."""
        # tandy can have VIEW PRINT 1 to 25, should raise IFC in that case
        error.throw_if(on and self.scroll_area.bottom == self.mode.height)
        self._bottom_bar.visible, was_visible = on, self._bottom_bar.visible
        if self._bottom_bar.visible != was_visible:
            self.redraw_bar()

    def redraw_bar(self):
        """Redraw bottom bar if visible, clear if not."""
        key_row = self.mode.height
        # Keys will only be visible on the active page at which KEY ON was given,
        # and only deleted on page at which KEY OFF given.
        self._apage.clear_rows(key_row, key_row, self._attr)
        if not self.mode.is_text_mode:
            reverse_attr = self._attr
        elif (self._attr >> 4) & 0x7 == 0:
            reverse_attr = 0x70
        else:
            reverse_attr = 0x07
        if self._bottom_bar.visible:
            with self.collect_updates():
                # always show only complete 8-character cells
                # this matters on pcjr/tandy width=20 mode
                for col in range((self.mode.width//8) * 8):
                    char, reverse = self._bottom_bar.get_char_reverse(col)
                    attr = reverse_attr if reverse else self._attr
                    self._apage.put_char_attr(key_row, col+1, char, attr)
            self.set_row_length(self.mode.height, self.mode.width)

    ###########################################################################
    # text retrieval on vpage (clipboard and print screen)
    # or apage (input, interactive commands)

    def get_chars(self, as_type=bytes):
        """Get all characters on the visible page, as tuple of tuples of bytes (raw) or unicode (dbcs)."""
        return self._pages[self._vpagenum].get_chars(as_type=as_type)

    def get_text(
            self,
            start_row=1, start_col=1, stop_row=None, stop_col=None,
            pagenum=None, wrap=True, as_type=text_type,
            stop_col_first=None
        ):
        """
        Retrieve consecutive rows of text on page `pagenum`,
        as tuple of bytes (raw) / tuple of unicode (dbcs).
        Each row is one (bytes or unicode) string and ends at the row length.
        So poked characters may not be included.
        Wrapped lines are output as a single row if `wrap=True`.
        """
        if pagenum is None:
            pagenum = self._vpagenum
        page = self._pages[pagenum]
        chars = page.get_chars(as_type)
        if stop_row is None:
            stop_row = len(chars)
        # get the rows of separate chars
        rows = [
            _charrow[:page.row_length(_row1)]
            for _row1, _charrow in enumerate(chars, start=1)
            if start_row <= _row1 <= stop_row
        ]
        # if a row is not complete and wraps, it has a \n
        if as_type == bytes:
            lf = (b'\n',)
        else:
            lf = ('\n',)
        rows = [
            _charrow + lf if page.has_linefeed(_row1) else _charrow
            for _row1, _charrow in enumerate(rows, start=start_row)
        ]
        if stop_col is None:
            stop_col = len(rows[-1])
        if stop_col_first is None:
            stop_col_first = len(rows[0])
        # apply start and stop column (inclusive bounds, base-1)
        if start_row == stop_row:
            rows[0] = rows[0][start_col-1:min(stop_col, stop_col_first)]
        else:
            # don't apply stop_col_first / furthest_right if we have multiple rows
            rows[0] = rows[0][start_col-1:]
            rows[-1] = rows[-1][:stop_col]
        # join the characters in each row
        output = tuple(as_type().join(_row) for _row in rows)
        if wrap:
            prev_wraps = [False] + list(
                page.wraps(_row)
                # don't include the stop_row as we need previous-row wraps anyway
                for _row in range(start_row, stop_row)
            )
            wrapped_output = []
            for row, prev_row_wrap in zip(output, prev_wraps):
                if prev_row_wrap:
                    wrapped_output[-1] += row
                else:
                    wrapped_output.append(row)
            output = tuple(wrapped_output)
        return output

    def get_logical_line(self, from_row, as_type=bytes, furthest_left=1, furthest_right=None):
        """Get the contents of the logical line on the active page, as one bytes string."""
        # find start and end of logical line
        rowtuple = self.get_text(
            start_row=self.find_start_of_line(from_row),
            start_col=furthest_left, stop_col_first=furthest_right,
            stop_row=self.find_end_of_line(from_row),
            pagenum=self._apagenum, as_type=as_type
        )
        return as_type().join(rowtuple)

    ###########################################################################
    # text screen callbacks

    def locate_(self, args):
        """LOCATE: Set cursor position, shape and visibility."""
        args = list(None if arg is None else values.to_int(arg) for arg in args)
        args = args + [None] * (5-len(args))
        row, col, cursor, start, stop = args
        row = self.current_row if row is None else row
        col = self.current_col if col is None else col
        error.throw_if(row == self.mode.height and self._bottom_bar.visible)
        if self.scroll_area.active:
            error.range_check(self.scroll_area.top, self.scroll_area.bottom, row)
        else:
            error.range_check(1, self.mode.height, row)
        error.range_check(1, self.mode.width, col)
        if row == self.mode.height:
            # temporarily allow writing on last row
            self._bottom_row_allowed = True
        self.set_pos(row, col, scroll_ok=False)
        if cursor is not None:
            error.range_check(0, (255 if self._tandytext else 1), cursor)
            # set cursor visibility - this should set the flag but have no effect in graphics modes
            self._cursor.set_textmode_override(cursor != 0)
        error.throw_if(start is None and stop is not None)
        if stop is None:
            stop = start
        if start is not None:
            error.range_check(0, 31, start, stop)
            # cursor shape only has an effect in text mode
            if self.mode.is_text_mode:
                self._cursor.set_shape(start, stop)

    def csrlin_(self, args):
        """CSRLIN: get the current screen row."""
        list(args)
        if (
                self.overflow and self.current_col == self.mode.width
                and self.current_row < self.scroll_area.bottom
            ):
            # in overflow position, return row+1 except on the last row
            csrlin = self.current_row + 1
        else:
            csrlin = self.current_row
        return self._values.new_integer().from_int(csrlin)

    def pos_(self, args):
        """POS: get the current screen column."""
        list(args)
        if self.current_col == self.mode.width and self.overflow:
            # in overflow position, return column 1.
            pos = 1
        else:
            pos = self.current_col
        return self._values.new_integer().from_int(pos)

    def screen_fn_(self, args):
        """SCREEN: get char or attribute at a location."""
        row = values.to_integer(next(args))
        col = values.to_integer(next(args))
        want_attr = next(args)
        if want_attr is not None:
            want_attr = values.to_integer(want_attr)
            want_attr = want_attr.to_int()
            error.range_check(0, 255, want_attr)
        row, col = row.to_int(), col.to_int()
        error.range_check(0, self.mode.height, row)
        error.range_check(0, self.mode.width, col)
        error.throw_if(row == 0 and col == 0)
        list(args)
        row = row or 1
        col = col or 1
        if self.scroll_area.active:
            error.range_check(self.scroll_area.top, self.scroll_area.bottom, row)
        if want_attr:
            if not self.mode.is_text_mode:
                result = 0
            else:
                result = self._apage.get_attr(row, col)
        else:
            result = self._apage.get_byte(row, col)
        return self._values.new_integer().from_int(result)

    def view_print_(self, args):
        """VIEW PRINT: set scroll region."""
        start, stop = (None if arg is None else values.to_int(arg) for arg in args)
        if start is None and stop is None:
            self.scroll_area.unset()
        else:
            if self._tandytext and not self._bottom_bar.visible:
                max_line = 25
            else:
                max_line = 24
            error.range_check(1, max_line, start, stop)
            error.throw_if(stop < start)
            self.scroll_area.set(start, stop)
            #set_pos(start, 1)
            self.overflow = False
            self.current_row, self.current_col = start, 1
            self._refresh_cursor()



"""
PC-BASIC - userfunctions.py
User-defined functions.

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import struct

from ...compat import int2byte, zip

from ..base import error
from ..base import codestream
from ..base import tokens as tk
from .. import values


class UserFunction(object):
    """User-defined function."""

    def __init__(self, name, code_stream, varnames, memory, expression_parser):
        """Define function."""
        self._codestream = code_stream
        self._start_loc = code_stream.tell()
        self._is_parsing = False
        self._memory = memory
        # if type not specified, it is evaluated at evaluation time, not at creation time
        self._varnames = varnames
        self._sigil = name[-1:]
        self._expression_parser = expression_parser

    def number_arguments(self):
        """Retrieve number of arguments."""
        return len(self._varnames)

    def evaluate(self, iargs):
        """Evaluate user-defined function."""
        # parse/evaluate arguments
        conversions = (values.TYPE_TO_CONV[self._memory.complete_name(name)[-1:]] for name in self._varnames)
        args = [conv(arg) for arg, conv in zip(iargs, conversions)]
        # recursion is not allowed as there's no way to terminate it
        if self._is_parsing:
            raise error.BASICError(error.OUT_OF_MEMORY)
        # parse/evaluate function expression
        # save existing vars
        varsave = {}
        varnames = [self._memory.complete_name(_v) for _v in self._varnames]
        for name in varnames:
            # set to 0 if they don't yet exist
            if name not in self._memory.scalars:
                self._memory.scalars.set(name)
            # copy the buffer
            varsave[name] = self._memory.scalars.view(name).clone()
        # set variables
        for name, value in zip(varnames, args):
            # append sigil, if missing
            name = self._memory.complete_name(name)
            self._memory.scalars.set(name, value)
        # set recursion flag
        self._is_parsing = True
        save_loc = self._codestream.tell()
        try:
            self._codestream.seek(self._start_loc)
            value = self._expression_parser.parse(self._codestream)
            return values.to_type(self._sigil, value)
        finally:
            self._codestream.seek(save_loc)
            # unset recursion flag
            self._is_parsing = False
            # restore existing vars, remove ones added
            for name in varsave:
                # re-assign the stored value
                self._memory.scalars.view(name).copy_from(varsave[name])



class UserFunctionManager(object):
    """User-defined function handler."""

    def __init__(self, memory, values, expression_parser):
        """Initialise functions."""
        self._fn_dict = {}
        # state variable for detecting recursion
        self._memory = memory
        self._values = values
        self._expression_parser = expression_parser

    def clear(self):
        """Clear all user-defined functions."""
        self._fn_dict.clear()

    def get(self, fnname):
        """Retrieve function by name."""
        # append sigil, if missing
        fnname = self._memory.complete_name(fnname)
        # parse/evaluate arguments
        try:
            fn = self._fn_dict[fnname]
        except KeyError:
            raise error.BASICError(error.UNDEFINED_USER_FUNCTION)
        if fn is None:
            raise error.BASICError(error.STX)
        return fn

    def define(self, fnname, ins):
        """Define a function."""
        ins.skip_blank()
        pointer_loc = self._memory.code_start + ins.tell()
        # read parameters
        fnvars = []
        if ins.skip_blank_read_if((b'(',)):
            while True:
                name = ins.read_name()
                # must not be empty
                error.throw_if(not name, error.STX)
                # do not append sigil here yet, leave until evaluation time
                fnvars.append(name)
                if ins.skip_blank() in tk.END_STATEMENT + (b')',):
                    break
                ins.require_read((b',',))
            ins.require_read((b')',))
        # read code
        if not ins.skip_blank_read_if((tk.O_EQ,)):
            self._fn_dict[fnname] = None
            return
        self._fn_dict[fnname] = UserFunction(fnname, ins, fnvars, self._memory, self._expression_parser)
        ins.skip_to(tk.END_STATEMENT)
        # update memory model
        # allocate function pointer
        pointer = struct.pack('<H', pointer_loc) + bytearray(values.size_bytes(fnname)-2)
        # function name is represented with first char shifted by 128
        memory_name = int2byte(128+ord(fnname[0:1])) + fnname[1:]
        self._memory.scalars.set(memory_name, self._values.from_bytes(pointer))
        for name in fnvars:
            # allocate, but don't set, variables
            # this is essential for them to be preserved during evaluate()
            name = self._memory.complete_name(name)
            self._memory.scalars.set(name)



"""
PC-BASIC - interface.video
Base class for video plugins

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import time

from ..compat import queue
from ..basic.base import signals


class VideoPlugin(object):
    """Base class for display/input interface plugins."""

    def __init__(self, input_queue, video_queue, **kwargs):
        """Setup the interface."""
        self.alive = True
        self.busy = False
        self._input_queue = input_queue
        self._video_queue = video_queue
        self._handlers = {
            signals.VIDEO_SET_MODE: self.set_mode,
            signals.VIDEO_UPDATE: self.update,
            signals.VIDEO_CLEAR_ROWS: self.clear_rows,
            signals.VIDEO_SCROLL: self.scroll,
            signals.VIDEO_SET_PALETTE: self.set_palette,
            signals.VIDEO_SET_CURSOR_SHAPE: self.set_cursor_shape,
            signals.VIDEO_SHOW_CURSOR: self.show_cursor,
            signals.VIDEO_MOVE_CURSOR: self.move_cursor,
            signals.VIDEO_SET_BORDER_ATTR: self.set_border_attr,
            signals.VIDEO_SET_CAPTION: self.set_caption_message,
            signals.VIDEO_SET_CLIPBOARD_TEXT: self.set_clipboard_text,
        }

    # called by Interface

    def cycle(self):
        """Video/input event cycle."""
        if self.alive:
            self._drain_queue()
        if self.alive:
            self._work()
            self._check_input()

    def sleep(self, ms):
        """Sleep a tick"""
        time.sleep(ms/1000.)

    # private methods

    def _drain_queue(self):
        """Drain signal queue."""
        while True:
            try:
                signal = self._video_queue.get(False)
            except queue.Empty:
                return True
            # putting task_done before the execution avoids hanging on join() after an exception
            self._video_queue.task_done()
            if signal.event_type == signals.QUIT:
                # close thread
                self.alive = False
            else:
                try:
                    self._handlers[signal.event_type](*signal.params)
                except KeyError:
                    pass

    # plugin overrides

    def __exit__(self, type, value, traceback):
        """Close the interface."""

    def __enter__(self):
        """Final initialisation."""
        return self

    def _work(self):
        """Display update cycle."""

    def _check_input(self):
        """Input devices update cycle."""

    # signal handlers

    def set_mode(self, canvas_height, canvas_width, text_height, text_width):
        """Initialise a given text or graphics mode."""

    def set_caption_message(self, msg):
        """Add a message to the window caption."""

    def set_clipboard_text(self, text):
        """Put text on the clipboard."""

    def set_palette(self, attributes, pack_pixels):
        """Build the palette."""

    def set_border_attr(self, attr):
        """Change the border attribute."""

    def clear_rows(self, back_attr, start_row, stop_row):
        """Clear a range of screen rows."""

    def show_cursor(self, cursor_on, cursor_blinks):
        """Change visibility of cursor."""

    def move_cursor(self, row, col, attr, width):
        """Move the cursor to a new position and set attribute and width."""

    def scroll(self, direction, start_row, stop_row, back_attr):
        """Scroll the screen between start_row and stop_row. direction 1 is down, -1 up."""

    def set_cursor_shape(self, from_line, to_line):
        """Build a sprite for the cursor."""

    def update(self, row, col, unicode_matrix, attr_matrix, y0, x0, sprite):
        """Put text or pixels at a given position."""
